#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_arch_prctl(ql, ARCHX, ARCH_SET_FS, *args, **kw):
    FSMSR = 0xC0000100
    ql.reg.msr(FSMSR, ARCH_SET_FS)
    regreturn = 0
    ql.nprint("arch_prctl(0x%x) = %d" % (ARCH_SET_FS, regreturn))
    ql.os.definesyscall_return(regreturn)


def ql_syscall_prctl(ql, *args, **kw):
    regreturn = 0
    ql.nprint("prctl() = %d" % (regreturn))
    ql.os.definesyscall_return(regreturn)



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

##############################################
# These are part of the core.py Qiling class #
# handling structure packing and unpacking   #
# for different architecture bits            #
##############################################

import struct
from .const import QL_ENDIAN
from .exception import QlErrorStructConversion

class QlCoreStructs(object):
    def __init__(self):
        super().__init__()
        self.archendian = None
        self.archbit = 0

    def pack64(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.pack('>Q', x)
        else:
            return struct.pack('Q', x)

    def pack64s(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.pack('>q', x)
        else:
            return struct.pack('q', x)

    def unpack64(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.unpack('>Q', x)[0]
        else:
            return struct.unpack('Q', x)[0]

    def unpack64s(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.unpack('>q', x)[0]
        else:
            return struct.unpack('q', x)[0]

    def pack32(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.pack('>I', x)
        else:
            return struct.pack('I', x)

    def pack32s(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.pack('>i', x)
        else:
            return struct.pack('i', x)

    def unpack32(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.unpack('>I', x)[0]
        else:
            return struct.unpack('I', x)[0]

    def unpack32s(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.unpack('>i', x)[0]
        else:
            return struct.unpack('i', x)[0]

    def pack16(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.pack('>H', x)
        else:
            return struct.pack('H', x)

    def pack16s(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.pack('>h', x)
        else:
            return struct.pack('h', x)

    def unpack16(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.unpack('>H', x)[0]
        else:
            return struct.unpack('H', x)[0]

    def unpack16s(self, x):
        if self.archendian == QL_ENDIAN.EB:
            return struct.unpack('>h', x)[0]
        else:
            return struct.unpack('h', x)[0]

    def pack(self, data):
        if self.archbit == 64:
            return self.pack64(data)
        elif self.archbit == 32:
            return self.pack32(data)
        elif self.archbit == 16:
            return self.pack16(data)
        raise QlErrorStructConversion("[!] Qiling pack conversion failed!")

    def packs(self, data):
        if self.archbit == 64:
            return self.pack64s(data)
        elif self.archbit == 32:
            return self.pack32s(data)
        elif self.archbit == 16:
            return self.pack16s(data)
        raise QlErrorStructConversion("[!] Qiling packs conversion failed!")

    def unpack(self, data):
        if self.archbit == 64:
            return self.unpack64(data)
        elif self.archbit == 32:
            return self.unpack32(data)
        elif self.archbit == 16:
            return self.unpack16(data)
        raise QlErrorStructConversion("[!] Qiling unpack conversion failed!")

    def unpacks(self, data):
        if self.archbit == 64:
            return self.unpack64s(data)
        elif self.archbit == 32:
            return self.unpack32s(data)
        elif self.archbit == 16:
            return self.unpack16s(data)
        raise QlErrorStructConversion("[!] Qiling unpacks conversion failed!")



from qiling.const import *
from qiling.os.const import *
from .const import *
from .utils import *
from .smm_base2_type import *
from .fncc import *


@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_SMM_BASE2_PROTOCOL)
    "InSmram": POINTER, #POINTER_T(ctypes.c_ubyte)
})
def hook_InSmm(ql, address, params):
    write_int64(ql, params["InSmram"], 0)
    return EFI_SUCCESS

@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_SMM_BASE2_PROTOCOL)
    "Smst": POINTER, #POINTER_T(POINTER_T(struct__EFI_SMM_SYSTEM_TABLE2))
})
def hook_GetSmstLocation(ql, address, params):
    if params["Smst"] == 0:
        return EFI_INVALID_PARAMETER
    write_int64(ql, params["Smst"], ql.loader.mm_system_table_ptr)
    return EFI_SUCCESS


# mm_system_table functions

@dxeapi(params={
    "Procedure": POINTER,
    "CpuNumber": INT, 
    "ProcArguments": POINTER
})
def hook_mm_startup_this_ap(ql, address, params):
    return EFI_INVALID_PARAMETER

@dxeapi(params={
    "HandlerType": GUID, 
    "Context": POINTER, 
    "CommBuffer": POINTER, 
    "CommBufferSize": POINTER, 
})
def hook_mm_interrupt_manage(ql, address, params):
    return EFI_NOT_FOUND

@dxeapi(params={
    "Handler": POINTER, 
    "HandlerType": GUID, 
    "DispatchHandle": POINTER, 
})
def hook_mm_interrupt_register(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "DispatchHandle": POINTER, 
})
def hook_efi_mm_interrupt_unregister(ql, address, params):
    return EFI_SUCCESS

def install_EFI_SMM_BASE2_PROTOCOL(ql, start_ptr, efi_mm_system_table):
    efi_smm_base2_protocol = EFI_SMM_BASE2_PROTOCOL()
    ptr = start_ptr
    pointer_size = 8
    efi_smm_base2_protocol.InSmm = ptr
    ql.hook_address(hook_InSmm, ptr)
    ptr += pointer_size
    efi_smm_base2_protocol.GetSmstLocation = ptr
    ql.hook_address(hook_GetSmstLocation, ptr)
    ptr += pointer_size

    # mm_system_table functions
    efi_mm_system_table.MmStartupThisAp = ptr
    ql.hook_address(hook_mm_startup_this_ap, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmiManage = ptr
    ql.hook_address(hook_mm_interrupt_manage, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmiHandlerRegister = ptr
    ql.hook_address(hook_mm_interrupt_register, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmiHandlerUnRegister = ptr
    ql.hook_address(hook_efi_mm_interrupt_unregister, ptr)
    ptr += pointer_size

    return (ptr, efi_smm_base2_protocol, efi_mm_system_table)




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 
import os


class Stat(object):
    def __init__(self, path):
        super().__init__()
        self.path = path

        self.st_dev = 0
        self.st_blksize = 0
        self.st_blocks = 0
        self.st_gid = 0
        self.st_ino = 0
        self.st_mode = 0
        self.st_nlink = 0
        self.st_rdev = 0
        self.st_size = 0
        self.st_uid = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0

        stat_buf = os.stat(self.path)
        for name in dir(stat_buf):
            if 'st_' in name:
                setattr(self, name, getattr(stat_buf, name))

class Fstat(object):
    def __init__(self, fd):
        super().__init__()
        self.fd = fd

        self.st_atime = 0
        self.st_blksize = 0
        self.st_blocks = 0
        self.st_ctime = 0
        self.st_dev = 0
        self.st_gid = 0
        self.st_ino = 0
        self.st_mode = 0
        self.st_mtime = 0
        self.st_nlink = 0
        self.st_rdev = 0
        self.st_size = 0
        self.st_uid = 0


        fstat_buf = os.fstat(self.fd)
        for name in dir(fstat_buf):
            if 'st_' in name:
                setattr(self, name, getattr(fstat_buf, name))


class Lstat(object):
    def __init__(self, path):
        super().__init__()
        self.path = path

        self.st_atime = 0
        self.st_blksize = 0
        self.st_blocks = 0
        self.st_ctime = 0
        self.st_dev = 0
        self.st_gid = 0
        self.st_ino = 0
        self.st_mode = 0
        self.st_mtime = 0
        self.st_nlink = 0
        self.st_rdev = 0
        self.st_size = 0
        self.st_uid = 0


        lstat_buf = os.lstat(self.path)
        for name in dir(lstat_buf):
            if 'st_' in name:
                setattr(self, name, getattr(lstat_buf, name))







#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

from qiling.const import *

linux_socket_types = {
    'SOCK_STREAM'    : 0x1,
    'SOCK_DGRAM'     : 0x2,
    'SOCK_RAW'       : 0x3,
    'SOCK_RDM'       : 0x4,
    'SOCK_SEQPACKET' : 0x5,
    'SOCK_PACKET'    : 0xa,
    'SOCK_NONBLOCK'  : 0x800,
    'SOCK_CLOEXEC'   : 0x80000,
}


linux_socket_domain = {
    'AF_UNSPEC'    : 0x0,
    'AF_LOCAL'     : 0x1,
    'AF_INET'      : 0x2,
    'AF_AX25'      : 0x3,
    'AF_IPX'       : 0x4,
    'AF_APPLETALK' : 0x5,
    'AF_NETROM'    : 0x6,
    'AF_BRIDGE'    : 0x7,
    'AF_AAL5'      : 0x8,
    'AF_X25'       : 0x9,
    'AF_INET6'     : 0xa,
    'AF_MAX'       : 0xc,
}

macos_socket_domain = {
    'AF_UNSPEC'    : 0x0,
    'AF_LOCAL'     : 0x1,
    'AF_INET'      : 0x2,
    'AF_IMPLINK'   : 0x3,
    'AF_PUP'       : 0x4,
    'AF_CHAOS'     : 0x5,
    'AF_NS'        : 0x6,
    'AF_ISO'       : 0x7,
    'AF_OSI'       : 0x7,
    'AF_ECMA'      : 0x8,
    'AF_DATAKIT'   : 0x9,
    'AF_CCITT'     : 0xa,
    'AF_SNA'       : 0xb,
    'AF_DECnet'    : 0xc,
    'AF_INET6'     : 0x1e,
}


arm_socket_types = {
    'SOCK_DGRAM'     : 0x1,
    'SOCK_STREAM'    : 0x2,
    'SOCK_RAW'       : 0x3,
    'SOCK_RDM'       : 0x4,
    'SOCK_SEQPACKET' : 0x5,
    'SOCK_DCCP'      : 0x6,
    'SOCK_PACKET'    : 0xa,
    'SOCK_NONBLOCK'  : 0x800,
    'SOCK_CLOEXEC'   : 0x80000,
}


arm_socket_domain = {
    'AF_UNSPEC'     : 0x0,
    'AF_FILE'       : 0x1,
    'AF_UNIX'       : 0x1,
    'AF_LOCAL'      : 0x1,
    'AF_INET'       : 0x2,
    'AF_AX25'       : 0x3,
    'AF_IPX'        : 0x4,
    'AF_APPLETALK'  : 0x5,
    'AF_NETROM'     : 0x6,
    'AF_BRIDGE'     : 0x7,
    'AF_ATMPVC'     : 0x8,
    'AF_X25'        : 0x9,
    'AF_INET6'      : 0xa,
    'AF_ROSE'       : 0xb,
    'AF_DECnet'     : 0xc,
    'AF_NETBEUI'    : 0xd,
    'AF_SECURITY'   : 0xe,
    'AF_KEY'        : 0xf,
    'AF_NETLINK'    : 0x10,
    'AF_ROUTE'      : 0x10,
    'AF_PACKET'     : 0x11,
    'AF_ASH'        : 0x12,
    'AF_ECONET'     : 0x13,
    'AF_ATMSVC'     : 0x14,
    'AF_RDS'        : 0x15,
    'AF_SNA'        : 0x16,
    'AF_IRDA'       : 0x17,
    'AF_PPPOX'      : 0x18,
    'AF_WANPIPE'    : 0x19,
    'AF_LLC'        : 0x1a,
    'AF_IB'         : 0x1b,
    'AF_MPLS'       : 0x1c,
    'AF_CAN'        : 0x1d,
    'AF_TIPC'       : 0x1e,
    'AF_BLUETOOTH'  : 0x1f,
    'AF_IUCV'       : 0x20,
    'AF_RXRPC'      : 0x21,
    'AF_ISDN'       : 0x22,
    'AF_PHONE'      : 0x23,
    'AF_IEEE802154' : 0x24,
    'AF_CAIF'       : 0x25,
    'AF_ALG'        : 0x26,
    'AF_NFC'        : 0x27,
    'AF_VSOCK'      : 0x28,
    'AF_KCM'        : 0x29,
    'AF_QIPCRTR'    : 0x2a,
    'AF_SMC'        : 0x2b,
    'AF_MAX'        : 0x2c,
}


mips_socket_types = {
    'SOCK_DGRAM'     : 0x1,
    'SOCK_STREAM'    : 0x2,
    'SOCK_RAW'       : 0x3,
    'SOCK_RDM'       : 0x4,
    'SOCK_SEQPACKET' : 0x5,
    'SOCK_DCCP'      : 0x6,
    'SOCK_PACKET'    : 0xa,
    'SOCK_CLOEXEC'   : 0x80000,
    'SOCK_NONBLOCK'  : 0x80,
}


mips_socket_domain = {
    'AF_UNSPEC'     : 0x0,
    'AF_FILE'       : 0x1,
    'AF_UNIX'       : 0x1,
    'AF_LOCAL'      : 0x1,
    'AF_INET'       : 0x2,
    'AF_AX25'       : 0x3,
    'AF_IPX'        : 0x4,
    'AF_APPLETALK'  : 0x5,
    'AF_NETROM'     : 0x6,
    'AF_BRIDGE'     : 0x7,
    'AF_ATMPVC'     : 0x8,
    'AF_X25'        : 0x9,
    'AF_INET6'      : 0xa,
    'AF_ROSE'       : 0xb,
    'AF_DECnet'     : 0xc,
    'AF_NETBEUI'    : 0xd,
    'AF_SECURITY'   : 0xe,
    'AF_KEY'        : 0xf,
    'AF_NETLINK'    : 0x10,
    'AF_ROUTE'      : 0x10,
    'AF_PACKET'     : 0x11,
    'AF_ASH'        : 0x12,
    'AF_ECONET'     : 0x13,
    'AF_ATMSVC'     : 0x14,
    'AF_RDS'        : 0x15,
    'AF_SNA'        : 0x16,
    'AF_IRDA'       : 0x17,
    'AF_PPPOX'      : 0x18,
    'AF_WANPIPE'    : 0x19,
    'AF_LLC'        : 0x1a,
    'AF_IB'         : 0x1b,
    'AF_MPLS'       : 0x1c,
    'AF_CAN'        : 0x1d,
    'AF_TIPC'       : 0x1e,
    'AF_BLUETOOTH'  : 0x1f,
    'AF_IUCV'       : 0x20,
    'AF_RXRPC'      : 0x21,
    'AF_ISDN'       : 0x22,
    'AF_PHONE'      : 0x23,
    'AF_IEEE802154' : 0x24,
    'AF_CAIF'       : 0x25,
    'AF_ALG'        : 0x26,
    'AF_NFC'        : 0x27,
    'AF_VSOCK'      : 0x28,
    'AF_KCM'        : 0x29,
    'AF_QIPCRTR'    : 0x2a,
    'AF_SMC'        : 0x2b,
    'AF_MAX'        : 0x2c,
}

mac_open_flags = {
    "O_RDONLY"   : 0x0000,
    "O_WRONLY"   : 0x0001,
    "O_RDWR"     : 0x0002,
    "O_NONBLOCK" : 0x0004,
    "O_APPEND"   : 0x0008,
    "O_ASYNC"    : 0x0040,
    "O_SYNC"     : 0x0080,
    "O_NOFOLLOW" : 0x0100,
    "O_CREAT"    : 0x0200,
    "O_TRUNC"    : 0x0400,
    "O_EXCL"     : 0x0800,
    "O_NOCTTY"   : 0x20000,
    "O_DIRECTORY": 0x100000
}


linux_open_flags = {
    'O_RDONLY'   : 0x0,
    'O_WRONLY'   : 0x1,
    'O_RDWR'     : 0x2,
    'O_CREAT'    : 0x40,
    'O_EXCL'     : 0x80,
    'O_NOCTTY'   : 0x100,
    'O_TRUNC'    : 0x200,
    'O_APPEND'   : 0x400,
    'O_NONBLOCK' : 0x800,
    'O_ASYNC'    : 0x2000,
    'O_DIRECTORY': 0x10000,
    'O_NOFOLLOW' : 0x20000,
    'O_SYNC'     : 0x101000,
}


mips_open_flags = {
    'O_RDONLY'   : 0x0,
    'O_WRONLY'   : 0x1,
    'O_RDWR'     : 0x2,
    'O_APPEND'   : 0x8,
    'O_NONBLOCK' : 0x80,
    'O_CREAT'    : 0x100,
    'O_TRUNC'    : 0x200,
    'O_EXCL'     : 0x400,
    'O_NOCTTY'   : 0x800,
    'O_ASYNC'    : 0x1000,
    'O_SYNC'     : 0x4000,
    'O_NOFOLLOW' : 0x20000,
    'O_DIRECTORY': 0x100000,
}


arm_open_flags = {
    'O_RDONLY'   : 0x0,
    'O_WRONLY'   : 0x1,
    'O_RDWR'     : 0x2,
    'O_CREAT'    : 0x40,
    'O_EXCL'     : 0x80,
    'O_NOCTTY'   : 0x100,
    'O_TRUNC'    : 0x200,
    'O_APPEND'   : 0x400,
    'O_NONBLOCK' : 0x800,
    'O_ASYNC'    : 0x2000,
    'O_DIRECTORY': 0x10000,
    'O_NOFOLLOW' : 0x20000,
    'O_SYNC'     : 0x101000,
}



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

import os, sys, types

from .utils import QlOsUtils
from .const import *
from .filestruct import ql_file
from .mapper import QlFsMapper

from qiling.const import *

from unicorn.x86_const import *

class QlOs(QlOsUtils):
    def __init__(self, ql):
        super(QlOs, self).__init__(ql)
        self.ql = ql
        self.ql.uc = self.ql.arch.init_uc
        self.fs_mapper = QlFsMapper(ql)
        self.child_processes = False
        self.thread_management = None
        self.profile = self.ql.profile
        self.current_path = self.profile.get("MISC", "current_path")
        self.exit_code = 0
        self.services = {}
        self.elf_mem_start = 0x0

        if "fileno" not in dir(sys.stdin) or "fileno" not in dir(sys.stdout) or "fileno" not in dir(sys.stderr):
            # IDAPython has some hack on standard io streams and thus they don't have corresponding fds.
            if "buffer" in dir(sys.stdin):
                self.stdin = sys.stdin.buffer
            else:
                self.stdin = sys.stdin
            if "buffer" in dir(sys.stdout):
                self.stdout = sys.stdout.buffer
            else:
                self.stdout = sys.stdout
            if "buffer" in dir(sys.stderr):
                self.stderr = sys.stderr.buffer
            else:
                self.stderr = sys.stderr
        else:
            self.stdin = ql_file('stdin', sys.stdin.fileno())
            self.stdout = ql_file('stdout', sys.stdout.fileno())
            self.stderr = ql_file('stderr', sys.stderr.fileno())

        if self.ql.stdin != 0:
            self.stdin = self.ql.stdin

        if self.ql.stdout != 0:
            self.stdout = self.ql.stdout

        if self.ql.stderr != 0:
            self.stderr = self.ql.stderr

        if self.ql.archbit == 32:
            EMU_END = 0x8fffffff
        elif self.ql.archbit == 64:
            EMU_END = 0xffffffffffffffff
        
        elif self.ql.archbit == 16:
            # 20bit address lane
            EMU_END = 0xfffff   
        
        # defult exit point
        self.exit_point = EMU_END

        if self.ql.shellcoder:
            self.shellcoder_ram_size = int(self.profile.get("SHELLCODER", "ram_size"), 16)
            # this shellcode entrypoint does not work for windows
            # windows shellcode entry point will comes from pe loader
            self.entry_point = int(self.profile.get("SHELLCODER", "entry_point"), 16)

        # We can save every syscall called
        self.syscalls = {}
        self.syscalls_counter = 0
        self.appeared_strings = {}
        self.setup_output()

    def save(self):
        return {}

    def restore(self, saved_state):
        pass


    def find_containing_image(self, pc):
        for image in self.ql.loader.images:
            if image.base <= pc < image.end:
                return image

    def emu_error(self):
        self.ql.nprint("\n")

        for reg in self.ql.reg.register_mapping:
            if isinstance(reg, str):
                REG_NAME = reg
                REG_VAL = self.ql.reg.read(reg)
                self.ql.nprint("[-] %s\t:\t 0x%x" % (REG_NAME, REG_VAL))

        self.ql.nprint("\n")
        self.ql.nprint("[+] PC = 0x%x" % (self.ql.reg.arch_pc), end="")
        containing_image = self.find_containing_image(self.ql.reg.arch_pc)
        if containing_image:
            offset = self.ql.reg.arch_pc - containing_image.base
            self.ql.nprint(" (%s+0x%x)" % (containing_image.path, offset))
        else:
            self.ql.nprint("\n")
        self.ql.mem.show_mapinfo()

        try:
            buf = self.ql.mem.read(self.ql.reg.arch_pc, 8)
            self.ql.nprint("[+] %r" % ([hex(_) for _ in buf]))

            self.ql.nprint("\n")
            self.disassembler(self.ql, self.ql.reg.arch_pc, 64)
        except:
            self.ql.nprint("[!] Error: PC(0x%x) Unreachable" % self.ql.reg.arch_pc)


    def _x86_set_args(self, args):
        for i in range(len(args)):
            # skip ret_addr
            self.ql.stack_write((i + 1) * 4, args[i])


    def _x8664_set_args(self, args):
        reg_list=None
        if self.ql.ostype == QL_OS.LINUX:
            reg_list = [UC_X86_REG_RDI, UC_X86_REG_RSI, UC_X86_REG_RDX, UC_X86_REG_R10, UC_X86_REG_R8, UC_X86_REG_R9]
        elif self.ql.ostype == QL_OS.WINDOWS:
            reg_list = [UC_X86_REG_RCX, UC_X86_REG_RDX, UC_X86_REG_R8, UC_X86_REG_R9]
        if reg_list!=None:
            for i in range(len(args)):
                self.ql.uc.reg_write(reg_list[i], args[i])


    def set_function_args(self, args):
        if self.ql.archtype == QL_ARCH.X86:   # 32bit
            self._x86_set_args(args)
        else:   # 64bit
            self._x8664_set_args(args)


    def _x86_get_params_by_index(self, index):
        # index starts from 0
        # skip ret_addr
        return self.ql.stack_read((index + 1) * 4)


    def _x8664_get_params_by_index(self, index):
        reg_list = ["rcx", "rdx", "r8", "r9"]
        if index < 4:
            return self.ql.reg.read(reg_list[index])

        index -= 4
        # skip ret_addr
        return self.ql.stack_read((index + 5) * 8)


    def get_param_by_index(self, index):
        if self.ql.archtype == QL_ARCH.X86:
            return self._x86_get_params_by_index(index)
        elif self.ql.archtype == QL_ARCH.X8664:
            return self._x8664_get_params_by_index(index)


    def _x86_get_args(self, number):
        arg_list = []
        for i in range(number):
            # skip ret_addr
            arg_list.append(self.ql.stack_read((i + 1) * 4))
        if number == 1:
            return arg_list[0]
        else:
            return arg_list


    def _x8664_get_args(self, number):
        reg_list = ["rcx", "rdx", "r8", "r9"]
        arg_list = []
        reg_num = number
        if reg_num > 4:
            reg_num = 4
        number -= reg_num
        for i in reg_list[:reg_num]:
            arg_list.append(self.ql.reg.read(i))
        for i in range(number):
            # skip ret_addr and 32 byte home space
            arg_list.append(self.ql.stack_read((i + 5) * 8))
        if reg_num == 1:
            return arg_list[0]
        else:
            return arg_list


    def set_function_params(self, in_params, out_params):
        index = 0
        for each in in_params:
            if in_params[each] == DWORD or in_params[each] == POINTER:
                out_params[each] = self.get_param_by_index(index)
            elif in_params[each] == ULONGLONG:
                if self.ql.archtype == QL_ARCH.X86:
                    low = self.get_param_by_index(index)
                    index += 1
                    high = self.get_param_by_index(index)
                    out_params[each] = high << 32 + low
                else:
                    out_params[each] = self.get_param_by_index(index)
            elif in_params[each] == STRING:
                ptr = self.get_param_by_index(index)
                if ptr == 0:
                    out_params[each] = 0
                else:
                    content = self.read_cstring(ptr)
                    out_params[each] = content
            elif in_params[each] == WSTRING:
                ptr = self.get_param_by_index(index)
                if ptr == 0:
                    out_params[each] = 0
                else:
                    content = self.read_wstring(ptr)
                    out_params[each] = content
            elif in_params[each] == GUID:
                ptr = self.get_param_by_index(index)
                if ptr == 0:
                    out_params[each] = 0
                else:
                    out_params[each] = str(self.read_guid(ptr))
            index += 1
        return index


    def get_function_param(self, number):
        if self.ql.archtype == QL_ARCH.X86:
            return self._x86_get_args(number)
        elif self.ql.archtype == QL_ARCH.X8664:
            return self._x8664_get_args(number)


    def set_return_value(self, ret):
        if self.ql.archtype == QL_ARCH.X86:
            self.ql.reg.eax = ret
        elif self.ql.archtype == QL_ARCH.X8664:
            self.ql.reg.rax = ret


    def get_return_value(self):
        if self.ql.archtype == QL_ARCH.X86:
            return self.ql.reg.eax
        elif self.ql.archtype == QL_ARCH.X8664:
            return self.ql.reg.rax

    #
    # stdcall cdecl fastcall cc
    #

    def __x86_cc(self, param_num, params, func, args, kwargs, passthru=False):
        # read params
        if params is not None:
            param_num = self.set_function_params(params, args[2])

        if isinstance(self.winapi_func_onenter, types.FunctionType):
            address, params = self.winapi_func_onenter(*args, **kwargs)
            args = (self.ql, address, params)
            onEnter = True
        else:
            onEnter = False

        # call function
        result = func(*args, **kwargs)

        if isinstance(self.winapi_func_onexit, types.FunctionType):
            self.winapi_func_onexit(*args, **kwargs)

        # set return value
        if result is not None:
            self.set_return_value(result)
        # print
        self.print_function(args[1], func.__name__, args[2], result, passthru)

        return result, param_num


    def _call_api(self, name, params, result, address, return_address):
        params_with_values = {}
        if name.startswith("hook_"):
            name = name.split("hook_", 1)[1]
            # printfs are shit
            if params is not None:
                self.set_function_params(params, params_with_values)
        self.syscalls.setdefault(name, []).append({
            "params": params_with_values,
            "result": result,
            "address": address,
            "return_address": return_address,
            "position": self.syscalls_counter
        })

        self.ql.os.syscalls_counter += 1


    def x86_stdcall(self, param_num, params, func, args, kwargs, passthru=False):
        # if we check ret_addr before the call, we can't modify the ret_addr from inside the hook
        result, param_num = self.__x86_cc(param_num, params, func, args, kwargs)

        # get ret addr
        ret_addr = self.ql.stack_read(0)

        # append syscall to list
        self._call_api(func.__name__, params, result, self.ql.reg.arch_pc, ret_addr)

        # update stack pointer
        self.ql.reg.arch_sp = self.ql.reg.arch_sp + ((param_num + 1) * 4)

        if not passthru and self.PE_RUN:
            self.ql.reg.arch_pc = ret_addr

        return result


    def x86_cdecl(self, param_num, params, func, args, kwargs, passthru=False):
        result, param_num = self.__x86_cc(param_num, params, func, args, kwargs)
        old_pc = self.ql.reg.arch_pc
        # append syscall to list
        self._call_api(func.__name__, params, result, old_pc, self.ql.stack_read(0))

        if not passthru and self.PE_RUN:
            self.ql.reg.arch_pc = self.ql.stack_pop()

        return result


    def x8664_fastcall(self, param_num, params, func, args, kwargs, passthru=False):
        result, param_num = self.__x86_cc(param_num, params, func, args, kwargs)

        old_pc = self.ql.reg.arch_pc
        # append syscall to list
        self._call_api(func.__name__, params, result, old_pc, self.ql.stack_read(0))

        if not passthru and self.PE_RUN:
           self.ql.reg.arch_pc = self.ql.stack_pop()

        return result



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *
from qiling.utils import *

def ql_syscall_wait4(ql, wait4_pid, wait4_wstatus, wait4_options, wait4_rusage, *args, **kw):
    spid, status, rusage = os.wait4(wait4_pid, wait4_options)
    if wait4_wstatus != 0:
        ql.mem.write(wait4_wstatus, ql.pack32(status))
    regreturn = spid
    ql.nprint("wait4(%d, %d) = %d"% (wait4_pid, wait4_options, regreturn))
    ql.os.definesyscall_return(regreturn)



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_writev(ql, writev_fd, writev_vec, writev_vien, *args, **kw):
    regreturn = 0
    size_t_len = ql.archbit // 8
    iov = ql.mem.read(writev_vec, writev_vien * size_t_len * 2)
    ql.nprint("writev(0x%x, 0x%x, 0x%x)" % (writev_fd, writev_vec, writev_vien))
    ql.dprint(D_CTNT, "[+] writev() CONTENT:")
    for i in range(writev_vien):
        addr = ql.unpack(iov[i * size_t_len * 2 : i * size_t_len * 2 + size_t_len])
        l = ql.unpack(iov[i * size_t_len * 2 + size_t_len : i * size_t_len * 2 + size_t_len * 2])
        regreturn += l
        ql.dprint(D_CTNT, "%s" % str(ql.mem.read(addr, l)))
    ql.os.definesyscall_return(regreturn)



# -*- coding: utf-8 -*-
#
# TARGET arch is: []
# WORD_SIZE is: 8
# POINTER_SIZE is: 8
# LONGDOUBLE_SIZE is: 16
#
import ctypes


c_int128 = ctypes.c_ubyte*16
c_uint128 = c_int128
void = None
if ctypes.sizeof(ctypes.c_longdouble) == 16:
    c_long_double_t = ctypes.c_longdouble
else:
    c_long_double_t = ctypes.c_ubyte*16

# if local wordsize is same as target, keep ctypes pointer function.
# required to access _ctypes
import _ctypes
# Emulate a pointer class using the approriate c_int32/c_int64 type
# The new class should have :
# ['__module__', 'from_param', '_type_', '__dict__', '__weakref__', '__doc__']
# but the class should be submitted to a unique instance for each base type
# to that if A == B, POINTER_T(A) == POINTER_T(B)
ctypes._pointer_t_type_cache = {}
def POINTER_T(pointee):
    # a pointer should have the same length as LONG
    fake_ptr_base_type = ctypes.c_uint64 
    # specific case for c_void_p
    if pointee is None: # VOID pointer type. c_void_p.
        pointee = type(None) # ctypes.c_void_p # ctypes.c_ulong
        clsname = 'c_void'
    else:
        clsname = pointee.__name__
    if clsname in ctypes._pointer_t_type_cache:
        return ctypes._pointer_t_type_cache[clsname]
    # make template
    _class = type('LP_%d_%s'%(8, clsname), (fake_ptr_base_type,),{}) 
    ctypes._pointer_t_type_cache[clsname] = _class
    return _class



undefined = ctypes.c_ubyte
ImageBaseOffset32 = ctypes.c_uint32
byte = ctypes.c_ubyte
dword = ctypes.c_uint32
longlong = ctypes.c_int64
qword = ctypes.c_uint64
uchar = ctypes.c_ubyte
uint = ctypes.c_uint32
ulonglong = ctypes.c_uint64
undefined1 = ctypes.c_ubyte
undefined2 = ctypes.c_uint16
undefined4 = ctypes.c_uint32
undefined8 = ctypes.c_uint64
ushort = ctypes.c_uint16
word = ctypes.c_uint16
class struct__EFI_SMM_SW_DISPATCH2_PROTOCOL(ctypes.Structure):
    pass

class struct_EFI_SMM_SW_REGISTER_CONTEXT(ctypes.Structure):
    pass

struct__EFI_SMM_SW_DISPATCH2_PROTOCOL._pack_ = True # source:False
struct__EFI_SMM_SW_DISPATCH2_PROTOCOL._functions_ = []
struct__EFI_SMM_SW_DISPATCH2_PROTOCOL._fields_ = [
    ('Register', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL), POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64))), POINTER_T(struct_EFI_SMM_SW_REGISTER_CONTEXT), POINTER_T(POINTER_T(None))))),
    ('UnRegister', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL), POINTER_T(None)))),
    ('MaximumSwiValue', ctypes.c_uint64),
]

struct__EFI_SMM_SW_DISPATCH2_PROTOCOL._functions_.append(("Register",['ctypes.c_uint64', 'POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL)', 'POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))', 'POINTER_T(struct_EFI_SMM_SW_REGISTER_CONTEXT)', 'POINTER_T(POINTER_T(None))']))
struct__EFI_SMM_SW_DISPATCH2_PROTOCOL._functions_.append(("UnRegister",['ctypes.c_uint64', 'POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL)', 'POINTER_T(None)']))
_EFI_SMM_SW_DISPATCH2_PROTOCOL = struct__EFI_SMM_SW_DISPATCH2_PROTOCOL
P_EFI_SMM_SW_DISPATCH2_PROTOCOL = POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL)
EFI_SMM_SW_DISPATCH2_PROTOCOL = struct__EFI_SMM_SW_DISPATCH2_PROTOCOL
UINT64 = ctypes.c_uint64
UINTN = ctypes.c_uint64
RETURN_STATUS = ctypes.c_uint64
EFI_STATUS = ctypes.c_uint64
EFI_HANDLE = POINTER_T(None)
EFI_MM_HANDLER_ENTRY_POINT = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))
EFI_SMM_HANDLER_ENTRY_POINT2 = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))
struct_EFI_SMM_SW_REGISTER_CONTEXT._pack_ = True # source:False
struct_EFI_SMM_SW_REGISTER_CONTEXT._functions_ = []
struct_EFI_SMM_SW_REGISTER_CONTEXT._fields_ = [
    ('SwSmiInputValue', ctypes.c_uint64),
]

EFI_SMM_SW_REGISTER_CONTEXT = struct_EFI_SMM_SW_REGISTER_CONTEXT
PEFI_SMM_SW_REGISTER_CONTEXT = POINTER_T(struct_EFI_SMM_SW_REGISTER_CONTEXT)
EFI_SMM_SW_REGISTER2 = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL), POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64))), POINTER_T(struct_EFI_SMM_SW_REGISTER_CONTEXT), POINTER_T(POINTER_T(None))))
EFI_SMM_SW_UNREGISTER2 = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL), POINTER_T(None)))
__all__ = \
    ['EFI_HANDLE', 'EFI_MM_HANDLER_ENTRY_POINT',
    'EFI_SMM_HANDLER_ENTRY_POINT2', 'EFI_SMM_SW_DISPATCH2_PROTOCOL',
    'EFI_SMM_SW_REGISTER2', 'EFI_SMM_SW_REGISTER_CONTEXT',
    'EFI_SMM_SW_UNREGISTER2', 'EFI_STATUS', 'ImageBaseOffset32',
    'PEFI_SMM_SW_REGISTER_CONTEXT', 'P_EFI_SMM_SW_DISPATCH2_PROTOCOL',
    'RETURN_STATUS', 'UINT64', 'UINTN',
    '_EFI_SMM_SW_DISPATCH2_PROTOCOL', 'byte', 'dword', 'longlong',
    'qword', 'struct_EFI_SMM_SW_REGISTER_CONTEXT',
    'struct__EFI_SMM_SW_DISPATCH2_PROTOCOL', 'uchar', 'uint',
    'ulonglong', 'undefined', 'undefined1', 'undefined2',
    'undefined4', 'undefined8', 'ushort', 'word']



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

import ctypes, logging, ntpath, os, pickle, platform

from .const import QL_ARCH_ENDIAN, QL_ENDIAN, QL_OS_POSIX, QL_OS_ALL, QL_OUTPUT, QL_OS
from .exception import QlErrorFileNotFound, QlErrorArch, QlErrorOsType, QlErrorOutput
from .utils import arch_convert, ostype_convert, output_convert
from .utils import ql_is_valid_arch, ql_get_arch_bits, verify_ret
from .utils import ql_setup_logging_env, ql_setup_logging_stream
from .core_struct import QlCoreStructs
from .core_hooks import QlCoreHooks
from .core_utils import QlCoreUtils
from .__version__ import __version__

class Qiling(QlCoreStructs, QlCoreHooks, QlCoreUtils):    
    def __init__(
            self,
            filename=None,
            rootfs=None,
            env=None,
            shellcoder=None,
            ostype=None,
            archtype=None,
            bigendian=False,
            output=None,
            verbose=1,
            profile=None,
            console=True,
            log_dir=None,
            log_split=None,
            append=None,
            libcache = False,
            stdin=0,
            stdout=0,
            stderr=0,
    ):
        super(Qiling, self).__init__()

        ##################################
        # Definition during ql=Qiling()  #
        ##################################
        self.filename = filename
        self.rootfs = rootfs
        self.env = env if env else {}
        self.shellcoder = shellcoder
        self.ostype = ostype
        self.archtype = archtype
        self.bigendian = bigendian
        self.output = output
        self.verbose = verbose
        self.profile = profile
        self.console = console
        self.log_dir = log_dir
        self.log_split = log_split
        self.append = append
        self.libcache = libcache
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        
        ##################################
        # Definition after ql=Qiling()   #
        ##################################
        self.patch_bin = []
        self.patch_lib = []
        self.patched_lib = []
        self.log_file_fd = None
        self.debug_stop = False
        self.internal_exception = None
        self.platform = platform.system()
        self.debugger = None
        # due to the instablity of multithreading, added a swtich for multithreading
        self.multithread = False
        self.root = False
        self.filter = None

        """
        Qiling Framework Core Engine
        """
        # shellcoder settings
        if self.shellcoder:
            if (self.ostype and type(self.ostype) == str) and (self.archtype and type(self.archtype) == str):
                self.ostype = ostype_convert(self.ostype.lower())
                self.archtype = arch_convert(self.archtype.lower())
                self.filename = ["qilingshellcoder"]
                if self.rootfs is None:
                    self.rootfs = "."
        # file check
        if self.shellcoder is None:
            if not os.path.exists(str(self.filename[0])):
                raise QlErrorFileNotFound("[!] Target binary not found")
            if not os.path.exists(self.rootfs):
                raise QlErrorFileNotFound("[!] Target rootfs not found")
        
        self.path = (str(self.filename[0]))
        self.argv = self.filename
        self.targetname = ntpath.basename(self.filename[0])

        ##########
        # Loader #
        ##########        
        self.loader = self.loader_setup()

        ############
        # setup    #
        ############           
        self.profile = self.profile_setup()
        if self.append == None:
            self.append = self.profile["MISC"]["append"]
        if self.log_dir == None:
            self.log_dir = self.profile["LOG"]["dir"]
        if self.log_split == None:            
            self.log_split =  self.profile.getboolean('LOG', 'split')

         # Log's configuration
        if self.log_dir != "" and type(self.log_dir) == str:
            _logger = ql_setup_logging_env(self)
            self.log_file_fd = _logger
        elif self.console == True:
            _logger = ql_setup_logging_stream(self)
            self.log_file_fd = _logger

        # qiling output method conversion
        if self.output and type(self.output) == str:
            # setter / getter for output
            self.output = self.output.lower()
            if self.output not in QL_OUTPUT:
                raise QlErrorOutput("[!] OUTPUT required: either 'default', 'disasm', 'debug', 'dump'")
            
        # check verbose, only can check after ouput being defined
        if type(self.verbose) != int or self.verbose > 99 and (self.verbose > 0 and self.output not in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP)):
            raise QlErrorOutput("[!] verbose required input as int and less than 99")
        
        ####################################
        # Set pointersize (32bit or 64bit) #
        ####################################
        self.archbit = ql_get_arch_bits(self.archtype)
        self.pointersize = (self.archbit // 8)  
        
        # Endian for shellcode needs to set manually
        if self.shellcoder:
            self.archendian = QL_ENDIAN.EL
            if self.bigendian == True and self.archtype in (QL_ARCH_ENDIAN):
                self.archendian = QL_ENDIAN.EB

        #############
        # Component #
        #############
        self.mem = self.component_setup("os", "memory")
        self.reg = self.component_setup("arch", "register")

        #####################################
        # Architecture and OS               #
        #####################################
        # Load architecture's and os module #
        #####################################
        self.arch = self.arch_setup()
        self.os = self.os_setup()

        # Run the loader
        self.loader.run()
        
        # Setup Outpt
        self.os.setup_output()



    # Emulate the binary from begin until @end, with timeout in @timeout and
    # number of emulated instructions in @count
    def run(self, begin=None, end=None, timeout=0, count=0):
        # replace the original entry point, exit point, timeout and count
        self.entry_point = begin
        self.exit_point = end
        self.timeout = timeout
        self.count = count

        # init debugger
        if self.debugger != False and self.debugger != None:
            self.debugger = self.debugger_setup()

        # patch binary
        self.__enable_bin_patch()

        # emulate the binary
        self.os.run()

        # run debugger
        if self.debugger != False and self.debugger != None:
            self.debugger.run()


    # patch code to memory address
    def patch(self, addr, code, file_name=b''):
        if file_name == b'':
            self.patch_bin.append((addr, code))
        else:
            self.patch_lib.append((addr, code, file_name.decode()))
    

    # ql.output var getter
    @property
    def output(self):
        return self._output


    # ql.output - output var setter eg. QL_OUTPUT.DEFAULT and etc
    @output.setter
    def output(self, output):
        self._output = output_convert(output)
    

    # ql.platform - platform var = host os getter eg. LINUX and etc
    @property
    def platform(self):
        return self._platform


    # ql.platform - platform var = host os setter eg. LINUX and etc
    @platform.setter
    def platform(self, value):
        self._platform = ostype_convert(value.lower())


    def __enable_bin_patch(self):
        for addr, code in self.patch_bin:
            self.mem.write(self.loader.load_address + addr, code)


    def enable_lib_patch(self):
        for addr, code, filename in self.patch_lib:
            try:
                self.mem.write(self.mem.get_lib_base(filename) + addr, code)
            except:
                raise RuntimeError("Fail to patch %s at address 0x%x" % (filename, addr))


    # save all qiling instance states
    def save(self, reg=True, mem=True, fd=False, cpu_context=False, os_context=False, loader=False, snapshot=None):
        saved_states = {}

        if reg == True:
            saved_states.update({"reg": self.reg.save()})

        if mem == True:
            saved_states.update({"mem": self.mem.save()})

        if fd == True: 
            saved_states.update({"fd": self.os.fd.save()})

        if cpu_context == True:
            saved_states.update({"cpu_context": self.arch.context_save()})

        if os_context == True:
            saved_states.update({"os_context": self.os.save()})
        
        if loader == True:
            saved_states.update({"loader": self.loader.save()})

        if snapshot != None:
            with open(snapshot, "wb") as save_state:
                pickle.dump(saved_states, save_state)
        else:
            return saved_states


    # restore states qiling instance from saved_states
    def restore(self, saved_states=None, snapshot=None):

        # snapshot will be ignored if saved_states is set
        if saved_states == None and snapshot != None:
            with open(snapshot, "rb") as load_state:
                saved_states = pickle.load(load_state)

        if "cpu_context" in saved_states:
            self.arch.context_restore(saved_states["cpu_context"])

        if "reg" in saved_states:
            self.reg.restore(saved_states["reg"])

        if "mem" in saved_states:
            self.mem.restore(saved_states["mem"])
        
        if "fd" in saved_states:
            self.os.fd.restore(saved_states["fd"])

        if "os_context" in saved_states:
            self.os.restore(saved_states["os_context"])
        
        if "loader" in saved_states:
            self.loader.restore(saved_states["loader"])


    # stop emulation
    def emu_stop(self):
        self.uc.emu_stop()


    # start emulation
    def emu_start(self, begin, end, timeout=0, count=0):
        self.uc.emu_start(begin, end, timeout, count)
        
        if self.internal_exception != None:
            raise self.internal_exception



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *
from qiling.os.stat import *

def ql_syscall_chmod(ql, filename, mode, null1, null2, null3, null4):
    regreturn = 0
    filename = ql.mem.string(filename)
    ql.nprint("chmod(%s,%d) = %d" % (filename, mode, regreturn))
    ql.os.definesyscall_return(regreturn)


def ql_syscall_fstatat64(ql, fstatat64_fd, fstatat64_fname, fstatat64_buf, fstatat64_flag, *args, **kw):
    fstatat64_fname = ql.mem.string(fstatat64_fname)

    real_path = ql.os.transform_to_real_path(fstatat64_fname)
    relative_path = ql.os.transform_to_relative_path(fstatat64_fname)

    regreturn = -1
    if os.path.exists(real_path) == True:
        fstat64_info = Stat(real_path)

        # struct stat is : 80 addr is : 0x4000811bc8
        # buf.st_dev offest 0 8 0
        # buf.st_ino offest 8 8 0
        # buf.st_mode offest 10 4 0
        # buf.st_nlink offest 14 4 0
        # buf.st_uid offest 18 4 0
        # buf.st_gid offest 1c 4 0
        # buf.st_rdev offest 20 8 0
        # buf.st_size offest 30 8 274886889936
        # buf.st_blksize offest 38 4 8461328
        # buf.st_blocks offest 40 8 274877909532
        # buf.st_atime offest 48 8 274886368336
        # buf.st_mtime offest 58 8 274877909472
        # buf.st_ctime offest 68 8 274886368336
        # buf.__glibc_reserved offest 78 8
        fstat64_buf = ql.pack64(fstat64_info.st_dev)
        fstat64_buf += ql.pack64(fstat64_info.st_ino)
        fstat64_buf += ql.pack32(fstat64_info.st_mode)
        fstat64_buf += ql.pack32(fstat64_info.st_nlink)
        fstat64_buf += ql.pack32(1000)
        fstat64_buf += ql.pack32(1000)
        fstat64_buf += ql.pack64(fstat64_info.st_rdev)
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(fstat64_info.st_size)
        fstat64_buf += ql.pack32(fstat64_info.st_blksize)
        fstat64_buf += ql.pack32(0)
        fstat64_buf += ql.pack64(fstat64_info.st_blocks)
        fstat64_buf += ql.pack64(int(fstat64_info.st_atime))
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(int(fstat64_info.st_mtime))
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(int(fstat64_info.st_ctime))
        fstat64_buf += ql.pack64(0)
        ql.mem.write(fstatat64_buf,fstat64_buf)
        regreturn = 0

    ql.nprint("fstatat64(0x%x, %s) = %d" % (fstatat64_fd, relative_path, regreturn))

    if regreturn == 0:
        ql.dprint(D_INFO, "[+] Directory Found: %s" % relative_path)
    else:
        ql.dprint(D_INFO, "[!] Directory Not Found: %s" % relative_path)

    ql.os.definesyscall_return(regreturn)


def ql_syscall_fstat64(ql, fstat64_fd, fstat64_add, *args, **kw):

    if ql.os.fd[fstat64_fd].fstat() == -1:
        regreturn = 0

    elif fstat64_fd < 256 and ql.os.fd[fstat64_fd] != 0:
        user_fileno = fstat64_fd
        fstat64_info = ql.os.fd[user_fileno].fstat()

        if ql.archtype== QL_ARCH.ARM64:
            # struct stat is : 80 addr is : 0x4000811bc8
            # buf.st_dev offest 0 8 0
            # buf.st_ino offest 8 8 0
            # buf.st_mode offest 10 4 0
            # buf.st_nlink offest 14 4 0
            # buf.st_uid offest 18 4 0
            # buf.st_gid offest 1c 4 0
            # buf.st_rdev offest 20 8 0
            # buf.st_size offest 30 8 274886889936
            # buf.st_blksize offest 38 4 8461328
            # buf.st_blocks offest 40 8 274877909532
            # buf.st_atime offest 48 8 274886368336
            # buf.st_mtime offest 58 8 274877909472
            # buf.st_ctime offest 68 8 274886368336
            # buf.__glibc_reserved offest 78 8
            fstat64_buf = ql.pack64(fstat64_info.st_dev)
            fstat64_buf += ql.pack64(fstat64_info.st_ino)
            fstat64_buf += ql.pack32(fstat64_info.st_mode)
            fstat64_buf += ql.pack32(fstat64_info.st_nlink)
            fstat64_buf += ql.pack32(ql.os.uid)
            fstat64_buf += ql.pack32(ql.os.gid)
            fstat64_buf += ql.pack64(fstat64_info.st_rdev)
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(fstat64_info.st_size)
            fstat64_buf += ql.pack32(fstat64_info.st_blksize)
            fstat64_buf += ql.pack32(0)
            fstat64_buf += ql.pack64(fstat64_info.st_blocks)
            fstat64_buf += ql.pack64(int(fstat64_info.st_atime))
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(int(fstat64_info.st_mtime))
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(int(fstat64_info.st_ctime))
            fstat64_buf += ql.pack64(0)
        elif ql.archtype == QL_ARCH.MIPS:
            # struct stat is : a0 addr is : 0x7fffedc0
            # buf.st_dev offest 0 4 2049
            # buf.st_ino offest 10 8 2400362
            # buf.st_mode offest 18 4 16893
            # buf.st_nlink offest 1c 4 5
            # buf.st_uid offest 20 4 1000
            # buf.st_gid offest 24 4 1000
            # buf.st_rdev offest 28 4 0
            # buf.st_size offest 38 8 0
            # buf.st_blksize offest 58 4 4096
            # buf.st_blocks offest 60 8 136
            # buf.st_atime offest 40 4 1586616689
            # buf.st_mtime offest 48 4 1586616689
            # buf.st_ctime offest 50 4 1586616689
            if ql.platform == QL_OS.MACOS:
                fstat64_buf = ql.pack32s(fstat64_info.st_dev)
            else:
                fstat64_buf = ql.pack32(fstat64_info.st_dev)
            fstat64_buf += b'\x00' * 12
            fstat64_buf += ql.pack64(fstat64_info.st_ino)
            fstat64_buf += ql.pack32(fstat64_info.st_mode)
            fstat64_buf += ql.pack32(fstat64_info.st_nlink)
            fstat64_buf += ql.pack32(ql.os.uid)
            fstat64_buf += ql.pack32(ql.os.gid)
            fstat64_buf += ql.pack32(fstat64_info.st_rdev)
            fstat64_buf += b'\x00' * 12
            fstat64_buf += ql.pack64(fstat64_info.st_size)
            fstat64_buf += ql.pack64(int(fstat64_info.st_atime))
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(int(fstat64_info.st_mtime))
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack64(int(fstat64_info.st_ctime))
            fstat64_buf += ql.pack64(0)
            fstat64_buf += ql.pack32(fstat64_info.st_blksize)
            fstat64_buf += ql.pack32(0)
            fstat64_buf += ql.pack64(fstat64_info.st_blocks)        
        elif ql.archtype == QL_ARCH.ARM:
            # pack fstatinfo
            if ql.platform == QL_OS.MACOS:
                fstat64_buf = ql.pack64s(fstat64_info.st_dev)
            else:
                fstat64_buf = ql.pack64(fstat64_info.st_dev)
            fstat64_buf += ql.pack32(0)
            fstat64_buf += ql.pack32(fstat64_info.st_ino)
            fstat64_buf += ql.pack32(fstat64_info.st_mode)
            fstat64_buf += ql.pack32(fstat64_info.st_nlink)
            fstat64_buf += ql.pack32(fstat64_info.st_uid)
            fstat64_buf += ql.pack32(fstat64_info.st_gid)
            fstat64_buf += ql.pack64(fstat64_info.st_rdev) #?? fstat_info.st_rdev
            fstat64_buf += ql.pack64(0) 
            fstat64_buf += ql.pack64(fstat64_info.st_size)
            fstat64_buf += ql.pack64(fstat64_info.st_blksize) #?? fstat_info.st_blksize
            fstat64_buf += ql.pack64(fstat64_info.st_blocks) #?? fstat_info.st_blocks
            fstat64_buf += ql.pack64(int(fstat64_info.st_atime))
            fstat64_buf += ql.pack64(int(fstat64_info.st_mtime))
            fstat64_buf += ql.pack64(int(fstat64_info.st_ctime))
            fstat64_buf += ql.pack64(fstat64_info.st_ino)

        else:
            # pack fstatinfo
            if ql.platform == QL_OS.MACOS:
                fstat64_buf = ql.pack64s(fstat64_info.st_dev)
            else:
                fstat64_buf = ql.pack64(fstat64_info.st_dev)
            fstat64_buf += ql.pack64(0x0000000300c30000)
            fstat64_buf += ql.pack32(fstat64_info.st_mode)
            fstat64_buf += ql.pack32(fstat64_info.st_nlink)
            fstat64_buf += ql.pack32(fstat64_info.st_uid)
            fstat64_buf += ql.pack32(fstat64_info.st_gid)
            fstat64_buf += ql.pack64(0x0000000000008800) #?? fstat_info.st_rdev
            fstat64_buf += ql.pack32(0xffffd257)
            fstat64_buf += ql.pack64(fstat64_info.st_size)
            fstat64_buf += ql.pack32(0x00000400) #?? fstat_info.st_blksize
            fstat64_buf += ql.pack64(0x0000000000000000) #?? fstat_info.st_blocks
            fstat64_buf += ql.pack64(int(fstat64_info.st_atime))
            fstat64_buf += ql.pack64(int(fstat64_info.st_mtime))
            fstat64_buf += ql.pack64(int(fstat64_info.st_ctime))
            fstat64_buf += ql.pack64(fstat64_info.st_ino)

        ql.mem.write(fstat64_add, fstat64_buf)
        regreturn = 0
    else:
        regreturn = -1

    ql.nprint("fstat64(%d, 0x%x) = %d" % (fstat64_fd, fstat64_add, regreturn))
    if regreturn == 0:
        ql.dprint(D_INFO, "[+] fstat64 write completed")
    else:
        ql.dprint(D_INFO, "[!] fstat64 read/write fail")
    ql.os.definesyscall_return(regreturn)


def ql_syscall_fstat(ql, fstat_fd, fstat_add, *args, **kw):

    if fstat_fd < 256 and ql.os.fd[fstat_fd] != 0 and hasattr(ql.os.fd[fstat_fd], "fstat"):
        user_fileno = fstat_fd
        fstat_info = ql.os.fd[user_fileno].fstat()

        if ql.archtype== QL_ARCH.MIPS:
            # pack fstatinfo
            fstat_buf = ql.pack32(fstat_info.st_dev)
            fstat_buf += ql.pack32(0) * 3
            fstat_buf += ql.pack32(fstat_info.st_ino)
            fstat_buf += ql.pack32(fstat_info.st_mode)
            fstat_buf += ql.pack32(fstat_info.st_nlink)
            fstat_buf += ql.pack32(fstat_info.st_uid)
            fstat_buf += ql.pack32(fstat_info.st_gid)
            fstat_buf += ql.pack32(fstat_info.st_rdev)
            fstat_buf += ql.pack32(0) * 2
            fstat_buf += ql.pack32(fstat_info.st_size)
            fstat_buf += ql.pack32(0)
            fstat_buf += ql.pack32(int(fstat_info.st_atime))
            fstat_buf += ql.pack32(0)
            fstat_buf += ql.pack32(int(fstat_info.st_mtime))
            fstat_buf += ql.pack32(0)
            fstat_buf += ql.pack32(int(fstat_info.st_ctime))
            fstat_buf += ql.pack32(0)
            fstat_buf += ql.pack32(fstat_info.st_blksize)
            fstat_buf += ql.pack32(fstat_info.st_blocks)
            fstat_buf = fstat_buf.ljust(0x90, b'\x00')
        elif ql.archtype== QL_ARCH.X8664:
            if ql.platform == QL_OS.MACOS:
                fstat_buf = ql.pack64s(fstat_info.st_dev)
            else:
                fstat_buf = ql.pack64(fstat_info.st_dev)
            fstat_buf += ql.pack(fstat_info.st_ino)
            fstat_buf += ql.pack64(fstat_info.st_nlink)
            fstat_buf += ql.pack32(fstat_info.st_mode)
            fstat_buf += ql.pack32(fstat_info.st_uid)
            fstat_buf += ql.pack32(fstat_info.st_gid)
            fstat_buf += ql.pack32(0)
            fstat_buf += ql.pack64(fstat_info.st_rdev)
            fstat_buf += ql.pack64(fstat_info.st_size)
            fstat_buf += ql.pack64(fstat_info.st_blksize)
            fstat_buf += ql.pack64(fstat_info.st_blocks)
            fstat_buf += ql.pack64(int(fstat_info.st_atime))
            fstat_buf += ql.pack64(0)
            fstat_buf += ql.pack64(int(fstat_info.st_mtime))
            fstat_buf += ql.pack64(0)
            fstat_buf += ql.pack64(int(fstat_info.st_ctime))
            fstat_buf += ql.pack64(0)
        elif ql.archtype== QL_ARCH.ARM64:
            # struct stat is : 80 addr is : 0x4000811bc8
            # buf.st_dev offest 0 8 0
            # buf.st_ino offest 8 8 0
            # buf.st_mode offest 10 4 0
            # buf.st_nlink offest 14 4 0
            # buf.st_uid offest 18 4 0
            # buf.st_gid offest 1c 4 0
            # buf.st_rdev offest 20 8 0
            # buf.st_size offest 30 8 274886889936
            # buf.st_blksize offest 38 4 8461328
            # buf.st_blocks offest 40 8 274877909532
            # buf.st_atime offest 48 8 274886368336
            # buf.st_mtime offest 58 8 274877909472
            # buf.st_ctime offest 68 8 274886368336
            # buf.__glibc_reserved offest 78 8
            if ql.platform == QL_OS.MACOS:
                fstat_buf = ql.pack64s(fstat_info.st_dev)
            else:
                fstat_buf = ql.pack64(fstat_info.st_dev)
            fstat_buf += ql.pack64(fstat_info.st_ino)
            fstat_buf += ql.pack32(fstat_info.st_mode)
            fstat_buf += ql.pack32(fstat_info.st_nlink)
            fstat_buf += ql.pack32(ql.os.uid)
            fstat_buf += ql.pack32(ql.os.gid)
            fstat_buf += ql.pack64(fstat_info.st_rdev)
            fstat_buf += ql.pack64(0)
            fstat_buf += ql.pack64(fstat_info.st_size)
            fstat_buf += ql.pack32(fstat_info.st_blksize)
            fstat_buf += ql.pack32(0)
            fstat_buf += ql.pack64(fstat_info.st_blocks)
            fstat_buf += ql.pack64(int(fstat_info.st_atime))
            fstat_buf += ql.pack64(0)
            fstat_buf += ql.pack64(int(fstat_info.st_mtime))
            fstat_buf += ql.pack64(0)
            fstat_buf += ql.pack64(int(fstat_info.st_ctime))
            fstat_buf += ql.pack64(0)
        else:
            # pack fstatinfo
            fstat_buf = ql.pack32(fstat_info.st_dev)
            fstat_buf += ql.pack(fstat_info.st_ino)
            fstat_buf += ql.pack32(fstat_info.st_mode)
            fstat_buf += ql.pack32(fstat_info.st_nlink)
            fstat_buf += ql.pack32(fstat_info.st_uid)
            fstat_buf += ql.pack32(fstat_info.st_gid)
            fstat_buf += ql.pack32(fstat_info.st_rdev)
            fstat_buf += ql.pack32(fstat_info.st_size)
            fstat_buf += ql.pack32(fstat_info.st_blksize)
            fstat_buf += ql.pack32(fstat_info.st_blocks)
            fstat_buf += ql.pack32(int(fstat_info.st_atime))
            fstat_buf += ql.pack32(int(fstat_info.st_mtime))
            fstat_buf += ql.pack32(int(fstat_info.st_ctime))

        ql.mem.write(fstat_add, fstat_buf)
        regreturn = 0
    else:
        regreturn = -1

    ql.nprint("fstat(%d, 0x%x) = %d" % (fstat_fd, fstat_add, regreturn))
    if regreturn == 0:
        ql.dprint(D_INFO, "[+] fstat write completed")
    else:
        ql.dprint(D_INFO, "[!] fstat read/write fail")
    ql.os.definesyscall_return(regreturn)


# int stat64(const char *pathname, struct stat64 *buf);
def ql_syscall_stat64(ql, stat64_pathname, stat64_buf_ptr, *args, **kw):
    stat64_file = (ql.mem.string(stat64_pathname))

    real_path = ql.os.transform_to_real_path(stat64_file)
    relative_path = ql.os.transform_to_relative_path(stat64_file)
    if os.path.exists(real_path) == False:
        regreturn = -1
    else:
        stat64_info = Stat(real_path)

        if ql.archtype== QL_ARCH.MIPS:
            # packfstatinfo
            # name offset size
            # struct stat is : a0
            # buf.st_dev offest 0 4
            # buf.st_ino offest 10 8
            # buf.st_mode offest 18 4
            # buf.st_nlink offest 1c 4
            # buf.st_uid offest 20 4
            # buf.st_gid offest 24 4
            # buf.st_rdev offest 28 4
            # buf.st_size offest 38 8
            # buf.st_blksize offest 58 4
            # buf.st_blocks offest 60 8
            # buf.st_atime offest 40 4
            # buf.st_mtime offest 48 4
            # buf.st_ctime offest 50 4
            stat64_buf = ql.pack32(stat64_info.st_dev)
            stat64_buf += ql.pack32(0) * 3
            stat64_buf += ql.pack64(stat64_info.st_ino)
            stat64_buf += ql.pack32(stat64_info.st_mode)
            stat64_buf += ql.pack32(stat64_info.st_nlink)
            stat64_buf += ql.pack32(1000)
            stat64_buf += ql.pack32(1000)
            stat64_buf += ql.pack32(stat64_info.st_rdev)
            stat64_buf += ql.pack32(0) * 3
            stat64_buf += ql.pack64(stat64_info.st_size)
            stat64_buf += ql.pack64(int(stat64_info.st_atime))
            stat64_buf += ql.pack64(int(stat64_info.st_mtime))
            stat64_buf += ql.pack64(int(stat64_info.st_ctime))
            stat64_buf += ql.pack32(stat64_info.st_blksize)
            stat64_buf += ql.pack32(0)
            stat64_buf += ql.pack64(stat64_info.st_blocks)
        else:
            # packfstatinfo
            stat64_buf = ql.pack64(stat64_info.st_dev)
            stat64_buf += ql.pack64(0x0000000300c30000)
            stat64_buf += ql.pack32(stat64_info.st_mode)
            stat64_buf += ql.pack32(stat64_info.st_nlink)
            stat64_buf += ql.pack32(stat64_info.st_uid)
            stat64_buf += ql.pack32(stat64_info.st_gid)
            stat64_buf += ql.pack64(0x0000000000008800) #?? fstat_info.st_rdev
            stat64_buf += ql.pack32(0xffffd257)
            stat64_buf += ql.pack64(stat64_info.st_size)
            stat64_buf += ql.pack32(0x00000400) #?? fstat_info.st_blksize
            stat64_buf += ql.pack64(0x0000000000000000) #?? fstat_info.st_blocks
            stat64_buf += ql.pack64(int(stat64_info.st_atime))
            stat64_buf += ql.pack64(int(stat64_info.st_mtime))
            stat64_buf += ql.pack64(int(stat64_info.st_ctime))
            stat64_buf += ql.pack64(stat64_info.st_ino)

        ql.mem.write(stat64_buf_ptr, stat64_buf)
        regreturn = 0

    ql.nprint("stat64(%s, 0x%x) = %d" % (relative_path, stat64_buf_ptr, regreturn))
    if regreturn == 0:
        ql.dprint(D_INFO, "[+] stat64 write completed")
    else:
        ql.dprint(D_INFO, "[!] stat64 read/write fail")
    ql.os.definesyscall_return(regreturn)


# int stat(const char *path, struct stat *buf);
def ql_syscall_stat(ql, stat_path, stat_buf_ptr, *args, **kw):
    stat_file = (ql.mem.string(stat_path))

    real_path = ql.os.transform_to_real_path(stat_file)
    relative_path = ql.os.transform_to_relative_path(stat_file)

    if os.path.exists(real_path) == False:
        regreturn = -1
    else:
        stat_info = Stat(real_path)

        if ql.archtype== QL_ARCH.MIPS:
            # pack fstatinfo
            stat_buf = ql.pack32(stat_info.st_dev)
            stat_buf += ql.pack32(0) * 3
            stat_buf += ql.pack32(stat_info.st_ino)
            stat_buf += ql.pack32(stat_info.st_mode)
            stat_buf += ql.pack32(stat_info.st_nlink)
            stat_buf += ql.pack32(stat_info.st_uid)
            stat_buf += ql.pack32(stat_info.st_gid)
            stat_buf += ql.pack32(stat_info.st_rdev)
            stat_buf += ql.pack32(0) * 2
            stat_buf += ql.pack32(stat_info.st_size)
            stat_buf += ql.pack32(0)
            stat_buf += ql.pack32(int(stat_info.st_atime))
            stat_buf += ql.pack32(0)
            stat_buf += ql.pack32(int(stat_info.st_mtime))
            stat_buf += ql.pack32(0)
            stat_buf += ql.pack32(int(stat_info.st_ctime))
            stat_buf += ql.pack32(0)
            stat_buf += ql.pack32(stat_info.st_blksize)
            stat_buf += ql.pack32(stat_info.st_blocks)
            stat_buf = stat_buf.ljust(0x90, b'\x00')
        else:
            # pack statinfo
            stat_buf = ql.pack32(stat_info.st_mode)
            stat_buf += ql.pack32(stat_info.st_ino)
            stat_buf += ql.pack32(stat_info.st_dev)
            stat_buf += ql.pack32(stat_info.st_rdev)
            stat_buf += ql.pack32(stat_info.st_nlink)
            stat_buf += ql.pack32(stat_info.st_size)
            stat_buf += ql.pack32(stat_info.st_size)
            stat_buf += ql.pack32(stat_info.st_size)
            stat_buf += ql.pack32(int(stat_info.st_atime))
            stat_buf += ql.pack32(int(stat_info.st_mtime))
            stat_buf += ql.pack32(int(stat_info.st_ctime))
            stat_buf += ql.pack32(stat_info.st_blksize)
            stat_buf += ql.pack32(stat_info.st_blocks)

        regreturn = 0
        ql.mem.write(stat_buf_ptr, stat_buf)

    ql.nprint("stat(%s, 0x%x) = %d" % (relative_path, stat_buf_ptr, regreturn))
    if regreturn == 0:
        ql.dprint(D_INFO, "[+] stat() write completed")
    else:
        ql.dprint(D_INFO, "[!] stat() read/write fail")
    ql.os.definesyscall_return(regreturn)


def ql_syscall_lstat(ql, lstat_path, lstat_buf_ptr, *args, **kw):
    lstat_file = (ql.mem.string(lstat_path))

    real_path = ql.os.transform_to_real_path(lstat_file)
    relative_path = ql.os.transform_to_relative_path(lstat_file)

    if os.path.exists(real_path) == False:
        regreturn = -1
    else:
        lstat_info = Lstat(real_path)

        if ql.archtype== QL_ARCH.MIPS:
            # pack fstatinfo
            lstat_buf = ql.pack32(lstat_info.st_dev)
            lstat_buf += ql.pack32(0) * 3
            lstat_buf += ql.pack32(lstat_info.st_ino)
            lstat_buf += ql.pack32(lstat_info.st_mode)
            lstat_buf += ql.pack32(lstat_info.st_nlink)
            lstat_buf += ql.pack32(lstat_info.st_uid)
            lstat_buf += ql.pack32(lstat_info.st_gid)
            lstat_buf += ql.pack32(lstat_info.st_rdev)
            lstat_buf += ql.pack32(0) * 2
            lstat_buf += ql.pack32(lstat_info.st_size)
            lstat_buf += ql.pack32(0)
            lstat_buf += ql.pack32(int(lstat_info.st_atime))
            lstat_buf += ql.pack32(0)
            lstat_buf += ql.pack32(int(lstat_info.st_mtime))
            lstat_buf += ql.pack32(0)
            lstat_buf += ql.pack32(int(lstat_info.st_ctime))
            lstat_buf += ql.pack32(0)
            lstat_buf += ql.pack32(lstat_info.st_blksize)
            lstat_buf += ql.pack32(lstat_info.st_blocks)
            lstat_buf = lstat_buf.ljust(0x90, b'\x00')
        else:
            # pack statinfo
            lstat_buf = ql.pack32(lstat_info.st_mode)
            lstat_buf += ql.pack32(lstat_info.st_ino)
            lstat_buf += ql.pack32(lstat_info.st_dev)
            lstat_buf += ql.pack32(lstat_info.st_rdev)
            lstat_buf += ql.pack32(lstat_info.st_nlink)
            lstat_buf += ql.pack32(lstat_info.st_size)
            lstat_buf += ql.pack32(lstat_info.st_size)
            lstat_buf += ql.pack32(lstat_info.st_size)
            lstat_buf += ql.pack32(int(lstat_info.st_atime))
            lstat_buf += ql.pack32(int(lstat_info.st_mtime))
            lstat_buf += ql.pack32(int(lstat_info.st_ctime))
            lstat_buf += ql.pack32(lstat_info.st_blksize)
            lstat_buf += ql.pack32(lstat_info.st_blocks)

        regreturn = 0
        ql.mem.write(lstat_buf_ptr, lstat_buf)

    ql.nprint("lstat(%s, 0x%x) = %d" % (relative_path, lstat_buf_ptr, regreturn))
    if regreturn == 0:
        ql.dprint(D_INFO, "[+] lstat() write completed")
    else:
        ql.dprint(D_INFO, "[!] lstat() read/write fail")
    ql.os.definesyscall_return(regreturn)

def ql_syscall_mknodat(ql, dirfd, pathname, mode, dev, *args, **kw):
    # fix me. dirfd(relative path) not implement.
    file_path = ql.mem.string(pathname)
    real_path = ql.os.transform_to_real_path(file_path)
    ql.nprint("mknodat(%d, %s, 0%o, %d)" % (dirfd, real_path, mode, dev))
    try:
        os.mknod(real_path, mode, dev)
        regreturn = 0
    except:
        regreturn = -1
    ql.os.definesyscall_return(regreturn)


def ql_syscall_mkdir(ql, pathname, mode, *args, **kw):
    file_path = ql.mem.string(pathname)
    real_path = ql.os.transform_to_real_path(file_path)
    ql.nprint("mkdir(%s, 0%o)" % (real_path, mode))
    try:
        if not os.path.exists(real_path):
            os.mkdir(real_path, mode)
        regreturn = 0
    except:
        regreturn = -1
    ql.os.definesyscall_return(regreturn)


def ql_syscall_umask(ql, mode, *args, **kw):
    oldmask = os.umask(mode)
    ql.nprint("umask(0%o) return oldmask 0%o" % (mode, oldmask))
    regreturn = oldmask
    ql.os.definesyscall_return(regreturn)



# -*- coding: utf-8 -*-
#
# TARGET arch is: []
# WORD_SIZE is: 8
# POINTER_SIZE is: 8
# LONGDOUBLE_SIZE is: 16
#
import ctypes


# if local wordsize is same as target, keep ctypes pointer function.
# required to access _ctypes
import _ctypes
# Emulate a pointer class using the approriate c_int32/c_int64 type
# The new class should have :
# ['__module__', 'from_param', '_type_', '__dict__', '__weakref__', '__doc__']
# but the class should be submitted to a unique instance for each base type
# to that if A == B, POINTER_T(A) == POINTER_T(B)
ctypes._pointer_t_type_cache = {}
def POINTER_T(pointee):
    # a pointer should have the same length as LONG
    fake_ptr_base_type = ctypes.c_uint64 
    # specific case for c_void_p
    if pointee is None: # VOID pointer type. c_void_p.
        pointee = type(None) # ctypes.c_void_p # ctypes.c_ulong
        clsname = 'c_void'
    else:
        clsname = pointee.__name__
    if clsname in ctypes._pointer_t_type_cache:
        return ctypes._pointer_t_type_cache[clsname]
    # make template
    _class = type('LP_%d_%s'%(8, clsname), (fake_ptr_base_type,),{}) 
    ctypes._pointer_t_type_cache[clsname] = _class
    return _class



undefined = ctypes.c_ubyte
ImageBaseOffset32 = ctypes.c_uint32
byte = ctypes.c_ubyte
dword = ctypes.c_uint32
longlong = ctypes.c_int64
qword = ctypes.c_uint64
uchar = ctypes.c_ubyte
uint = ctypes.c_uint32
ulonglong = ctypes.c_uint64
undefined1 = ctypes.c_ubyte
undefined2 = ctypes.c_uint16
undefined4 = ctypes.c_uint32
undefined8 = ctypes.c_uint64
ushort = ctypes.c_uint16
word = ctypes.c_uint16
class struct__EFI_MM_ACCESS_PROTOCOL(ctypes.Structure):
    pass

class struct_EFI_MMRAM_DESCRIPTOR(ctypes.Structure):
    pass

struct__EFI_MM_ACCESS_PROTOCOL._pack_ = True # source:False
struct__EFI_MM_ACCESS_PROTOCOL._functions_ = []
struct__EFI_MM_ACCESS_PROTOCOL._fields_ = [
    ('Open', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)))),
    ('Close', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)))),
    ('Lock', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)))),
    ('GetCapabilities', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL), POINTER_T(ctypes.c_uint64), POINTER_T(struct_EFI_MMRAM_DESCRIPTOR)))),
    ('LockState', ctypes.c_ubyte),
    ('OpenState', ctypes.c_ubyte),
    ('PADDING_0', ctypes.c_ubyte * 6),
]

struct__EFI_MM_ACCESS_PROTOCOL._functions_.append(("Open",['ctypes.c_uint64', 'POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)']))
struct__EFI_MM_ACCESS_PROTOCOL._functions_.append(("Close",['ctypes.c_uint64', 'POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)']))
struct__EFI_MM_ACCESS_PROTOCOL._functions_.append(("Lock",['ctypes.c_uint64', 'POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)']))
struct__EFI_MM_ACCESS_PROTOCOL._functions_.append(("GetCapabilities",['ctypes.c_uint64', 'POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)', 'POINTER_T(ctypes.c_uint64)', 'POINTER_T(struct_EFI_MMRAM_DESCRIPTOR)']))
_EFI_MM_ACCESS_PROTOCOL = struct__EFI_MM_ACCESS_PROTOCOL
P_EFI_MM_ACCESS_PROTOCOL = POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)
UINT64 = ctypes.c_uint64
UINTN = ctypes.c_uint64
RETURN_STATUS = ctypes.c_uint64
EFI_STATUS = ctypes.c_uint64
EFI_MM_ACCESS_PROTOCOL = struct__EFI_MM_ACCESS_PROTOCOL
EFI_MM_OPEN = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)))
EFI_MM_CLOSE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)))
EFI_MM_LOCK = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)))
struct_EFI_MMRAM_DESCRIPTOR._pack_ = True # source:False
struct_EFI_MMRAM_DESCRIPTOR._functions_ = []
struct_EFI_MMRAM_DESCRIPTOR._fields_ = [
    ('PhysicalStart', ctypes.c_uint64),
    ('CpuStart', ctypes.c_uint64),
    ('PhysicalSize', ctypes.c_uint64),
    ('RegionState', ctypes.c_uint64),
]

EFI_MMRAM_DESCRIPTOR = struct_EFI_MMRAM_DESCRIPTOR
PEFI_MMRAM_DESCRIPTOR = POINTER_T(struct_EFI_MMRAM_DESCRIPTOR)
EFI_MM_CAPABILITIES = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL), POINTER_T(ctypes.c_uint64), POINTER_T(struct_EFI_MMRAM_DESCRIPTOR)))
BOOLEAN = ctypes.c_ubyte
EFI_PHYSICAL_ADDRESS = ctypes.c_uint64
__all__ = \
    ['BOOLEAN', 'EFI_MMRAM_DESCRIPTOR', 'EFI_MM_ACCESS_PROTOCOL',
    'EFI_MM_CAPABILITIES', 'EFI_MM_CLOSE', 'EFI_MM_LOCK',
    'EFI_MM_OPEN', 'EFI_PHYSICAL_ADDRESS', 'EFI_STATUS',
    'ImageBaseOffset32', 'PEFI_MMRAM_DESCRIPTOR',
    'P_EFI_MM_ACCESS_PROTOCOL', 'RETURN_STATUS', 'UINT64', 'UINTN',
    '_EFI_MM_ACCESS_PROTOCOL', 'byte', 'dword', 'longlong', 'qword',
    'struct_EFI_MMRAM_DESCRIPTOR', 'struct__EFI_MM_ACCESS_PROTOCOL',
    'uchar', 'uint', 'ulonglong', 'undefined', 'undefined1',
    'undefined2', 'undefined4', 'undefined8', 'ushort', 'word']



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

import os

from qiling.const import *

def ql_syscall_getrandom(ql, buf, buflen, flags,*args, **kw):
    data = None
    regreturn = None
    try:
        data = os.urandom(buflen)
        ql.uc.mem_write(buf, data)
        regreturn = len(data)
    except:
        regreturn = -1

    ql.nprint("getrandom(0x%x, 0x%x, 0x%x) = %d" %
              (buf, buflen, flags, regreturn))

    if data:
        ql.dprint(D_CTNT, "[+] getrandom() CONTENT:")
        ql.dprint(D_CTNT, str(data))
    ql.os.definesyscall_return(regreturn)



#!/usr/bin/env python3

import math, copy, os
from contextlib import contextmanager

from qiling.const import QL_ARCH

from .utils import dump_regs, get_arm_flags



# class for colorful prints
class color:
   CYAN      = '\033[96m'
   PURPLE    = '\033[95m'
   BLUE      = '\033[94m'
   YELLOW    = '\033[93m'
   GREEN     = '\033[92m'
   RED       = '\033[91m'
   DARKGRAY  = '\033[90m'
   WHITE     = '\033[48m'
   DARKCYAN  = '\033[36m'
   BLACK     = '\033[35m'
   UNDERLINE = '\033[4m'
   BOLD      = '\033[1m'
   END       = '\033[0m'


# read data from memory of qiling instance
def examine_mem(ql, xaddr, count):

    lines = 1 if count <= 4 else math.ceil(count / 4)

    mem_read = [ql.mem.read(xaddr+(offset*4), 4) for offset in range(count)]

    for line in range(lines):
        offset = line * 0x10
        print("0x%08x:\t" % (xaddr+offset), end="")

        idx = line * 4
        for each in mem_read[idx:idx+4]:
            print("0x%08x\t" % (ql.unpack(each)), end="")

        print()

# get terminal window height and width
def get_terminal_size():
    return map(int, os.popen('stty size', 'r').read().split())


# try to read data from ql memory
def _try_read(ql, address, size):
    try:
        result = ql.mem.read(address, size)
    except:
        result = None

    return result


# divider printer
@contextmanager
def context_printer(ql, field_name, ruler="="):
    _height, _width = get_terminal_size()
    print(field_name, ruler * (_width - len(field_name) - 1))
    yield
    print(ruler * _width)


def context_reg(ql, saved_states=None, *args, **kwargs):

    # context render for registers
    with context_printer(ql, "[Registers]"):

        _cur_regs = dump_regs(ql)

        _colors = (color.DARKCYAN, color.BLUE, color.RED, color.YELLOW, color.GREEN, color.PURPLE, color.CYAN, color.WHITE)

        if ql.archtype == QL_ARCH.MIPS:

            _cur_regs.update({"fp": _cur_regs.pop("s8")})

            if saved_states is not None:
                _saved_states = copy.deepcopy(saved_states)
                _saved_states.update({"fp": _saved_states.pop("s8")})
                _diff = [k for k in _cur_regs if _cur_regs[k] != _saved_states[k]]

            else:
                _diff = None

            lines = ""
            for idx, r in enumerate(_cur_regs, 1):
                line = "{}{}: 0x{{:08x}} {}\t".format(_colors[(idx-1) // 4], r, color.END)

                if _diff and r in _diff:
                    line = "{}{}".format(color.UNDERLINE, color.BOLD) + line

                if idx % 4 == 0 and idx != 32:
                    line += "\n"

                lines += line

            print(lines.format(*_cur_regs.values()))

        elif ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB):

            _cur_regs.update({"sl": _cur_regs.pop("r10")})
            _cur_regs.update({"fp": _cur_regs.pop("r11")})
            _cur_regs.update({"ip": _cur_regs.pop("r12")})

            if saved_states is not None:
                _saved_states = copy.deepcopy(saved_states)
                _saved_states.update({"sl": _saved_states.pop("r10")})
                _saved_states.update({"fp": _saved_states.pop("r11")})
                _saved_states.update({"ip": _saved_states.pop("r12")})
                _diff = [k for k in _cur_regs if _cur_regs[k] != _saved_states[k]]

            else:
                _diff = None

            lines = ""
            for idx, r in enumerate(_cur_regs, 1):
                line = "{}{:}: 0x{{:08x}} {}\t".format(_colors[(idx-1) // 4], r, color.END)

                if _diff and r in _diff:
                    line = "{}{}".format(color.UNDERLINE, color.BOLD) + line

                if idx % 4 == 0:
                    line += "\n"

                lines += line

            print(lines.format(*_cur_regs.values()))
            print(color.GREEN, "[{cpsr[mode]} mode], Thumb: {cpsr[thumb]}, FIQ: {cpsr[fiq]}, IRQ: {cpsr[irq]}, NEG: {cpsr[neg]}, ZERO: {cpsr[zero]}, Carry: {cpsr[carry]}, Overflow: {cpsr[overflow]}".format(cpsr=get_arm_flags(ql.reg.cpsr)), color.END, sep="")

    # context render for Stack
    with context_printer(ql, "[Stack]", ruler="-"):

        for idx in range(8):
            _addr = ql.reg.arch_sp + idx * 4
            _val = ql.mem.read(_addr, ql.archbit // 8)
            print("$sp+0x%02x|[0x%08x]=> 0x%08x" % (idx*4, _addr, ql.unpack(_val)), end="")

            try: # try to deference wether its a pointer
                _deref = ql.mem.read(_addr, 4)
            except:
                _deref = None

            if _deref:
                print(" => 0x%08x" % ql.unpack(_deref))


def print_asm(ql, instructions):
    for ins in instructions:
        fmt = (ins.address, ins.mnemonic.ljust(6), ins.op_str)
        if ql.reg.arch_pc == ins.address:
            print("PC ==>  0x%x\t%s %s" % fmt)
        else:
            print("\t0x%x\t%s %s" % fmt)


def context_asm(ql, address, size, *args, **kwargs):

    with context_printer(ql, field_name="[Code]"):
        md = ql.create_disassembler()

        # assembly before current location

        pre_tmp = _try_read(ql, address-0x10, 0x10)
        if pre_tmp:
            pre_ins = md.disasm(pre_tmp, address-0x10)
            print_asm(ql, pre_ins)

        # assembly for current locaton

        tmp = ql.mem.read(address, size)
        cur_ins = md.disasm(tmp, address)
        print_asm(ql, cur_ins)

        # assembly after current locaton

        pos_tmp = _try_read(ql, address+4, 0x10)
        if pos_tmp:
            pos_ins = md.disasm(pos_tmp, address+4)
            print_asm(ql, pos_ins)



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

import magic, os

import pefile

from qiling.const import QL_OS, QL_OS_ALL, QL_ARCH, QL_ENDIAN
from qiling.exception import QlErrorArch, QlErrorOsType

def ql_checkostype(path):

    if os.path.isdir(path) and (str(path)).endswith(".kext"):
        return QL_ARCH.X8664, QL_OS.MACOS, QL_ENDIAN.EL

    arch = None
    ostype = None
    archendian = None

    ftype = magic.from_file(path)

    if "ELF" in ftype:
        arch, ostype, archendian = ql_elf_check_archtype(path)
    elif "Mach-O" in ftype:
        arch, ostype, archendian = ql_macho_check_archtype(path)
    elif "PE32" in ftype:
        arch, ostype, archendian = ql_pe_check_archtype(path)
    elif ("COM" in ftype and "DOS" in ftype) or "COM" in path:
        arch = QL_ARCH.A8086
        ostype = QL_OS.DOS
        archendian = QL_ENDIAN.EL
    elif "MBR" in ftype and "DOS" in ftype: # Yes, we put MBR into dos.py.
        arch = QL_ARCH.A8086
        ostype = QL_OS.DOS
        archendian = QL_ENDIAN.EL
    elif "MS-DOS" in ftype:
        # Here we have to distinguish between real 16bit DOS executables and EFI excutables.
        # I could confirm from specs that all UEFI executables should be PE/PE32+.
        # But 16bit DOS executables don't have a valid NT header.
        # I'm not sure why libmagic(file) classify EFI executables as "MS-DOS executable"
        try:
            pefile.PE(path)
        except pefile.PEFormatError:
            arch = QL_ARCH.A8086
            ostype = QL_OS.DOS
            archendian = QL_ENDIAN.EL
        else:
            arch, ostype, archendian = ql_pe_check_archtype(path)

    if ostype not in (QL_OS_ALL):
        raise QlErrorOsType("[!] File does not belong to either 'linux', 'windows', 'freebsd', 'macos', 'ios', 'dos'")

    return arch, ostype, archendian

def ql_elf_check_archtype(path):
    def getident():
        return elfdata

    with open(path, "rb") as f:
        elfdata = f.read()[:20]

    ident = getident()
    ostype = None
    arch = None
    archendian = None

    if ident[: 4] == b'\x7fELF':
        elfbit = ident[0x4]
        endian = ident[0x5]
        osabi = ident[0x7]
        e_machine = ident[0x12:0x14]

        if osabi == 0x09:
            ostype = QL_OS.FREEBSD
        elif osabi in (0x0, 0x03) or osabi >= 0x11:
            ostype = QL_OS.LINUX

        if e_machine == b"\x03\x00":
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.X86
        elif e_machine == b"\x08\x00" and endian == 1 and elfbit == 1:
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.MIPS
        elif e_machine == b"\x00\x08" and endian == 2 and elfbit == 1:
            archendian = QL_ENDIAN.EB
            arch = QL_ARCH.MIPS
        elif e_machine == b"\x28\x00" and endian == 1 and elfbit == 1:
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.ARM
        elif e_machine == b"\x00\x28" and endian == 2 and elfbit == 1:
            archendian = QL_ENDIAN.EB
            arch = QL_ARCH.ARM            
        elif e_machine == b"\xB7\x00":
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.ARM64
        elif e_machine == b"\x3E\x00":
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.X8664
        else:
            arch = None

    return arch, ostype, archendian

def ql_macho_check_archtype(path):
   
    def getident():
        return machodata

    with open(path, "rb") as f:
        machodata = f.read()[:32]

    ident = getident()

    macho_macos_sig64 = b'\xcf\xfa\xed\xfe'
    macho_macos_sig32 = b'\xce\xfa\xed\xfe'
    macho_macos_fat = b'\xca\xfe\xba\xbe'  # should be header for FAT

    ostype = None
    arch = None
    archendian = None

    if ident[: 4] in (macho_macos_sig32, macho_macos_sig64, macho_macos_fat):
        ostype = QL_OS.MACOS
    else:
        ostype = None

    if ostype:
        # if ident[0x7] == 0: # 32 bit
        #    arch = QL_ARCH.X86
        if ident[0x4] == 7 and ident[0x7] == 1:  # X86 64 bit
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.X8664
        elif ident[0x4] == 12 and ident[0x7] == 1:  # ARM64  ident[0x4] = 0x0C
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.ARM64
        else:
            arch = None

    return arch, ostype, archendian

def ql_pe_check_archtype(path):

    pe = pefile.PE(path, fast_load=True)
    ostype = None
    arch = None
    archendian = None

    machine_map = {
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_I386']: QL_ARCH.X86,
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_AMD64']: QL_ARCH.X8664,
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_ARM']: QL_ARCH.ARM,
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_THUMB']: QL_ARCH.ARM,
        # pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_ARM64']     :   QL_ARCH.ARM64       #pefile does not have the definition
        # for IMAGE_FILE_MACHINE_ARM64
        0xAA64: QL_ARCH.ARM64  # Temporary workaround for Issues #21 till pefile gets updated
    }
    # get arch
    archendian = QL_ENDIAN.EL
    arch = machine_map.get(pe.FILE_HEADER.Machine)

    if arch:
        if pe.OPTIONAL_HEADER.Subsystem >= pefile.SUBSYSTEM_TYPE['IMAGE_SUBSYSTEM_EFI_APPLICATION'] and \
        pe.OPTIONAL_HEADER.Subsystem <= pefile.SUBSYSTEM_TYPE['IMAGE_SUBSYSTEM_EFI_ROM'] :
            ostype = QL_OS.UEFI
        else:
            ostype = QL_OS.WINDOWS
    else:
        ostype = None

    return arch, ostype, archendian



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

import ctypes
import types
import struct
from unicorn import *
from unicorn.x86_const import *
from qiling.const import *
from qiling.os.os import QlOs

class QlOsUefi(QlOs):
    def __init__(self, ql):
        super(QlOsUefi, self).__init__(ql)
        self.ql = ql
        self.entry_point = 0
        self.running_module = None
        self.user_defined_api = {}
        self.user_defined_api_onenter = {}
        self.user_defined_api_onexit = {}
        self.PE_RUN = True
        self.heap = None # Will be initialized by the loader.
    
    def save(self):
        saved_state = super(QlOsUefi, self).save()
        saved_state['entry_point'] = self.entry_point
        return saved_state

    def restore(self, saved_state):
        super(QlOsUefi, self).restore(saved_state)
        self.entry_point = saved_state['entry_point']

    @staticmethod
    def notify_after_module_execution(ql, number_of_modules_left):
        return False
    
    @staticmethod
    def notify_before_module_execution(ql, module):
        ql.os.running_module = module
        return False

    def run(self):
        self.notify_before_module_execution(self.ql, self.running_module)
        if self.ql.exit_point is not None:
            self.exit_point = self.ql.exit_point
        
        if  self.ql.entry_point  is not None:
            self.ql.loader.entry_point = self.ql.entry_point

        try:
            self.ql.emu_start(self.ql.loader.entry_point, self.exit_point, self.ql.timeout, self.ql.count)
        except UcError:
            self.emu_error()
            raise

        if self.ql.internal_exception is not None:
            raise self.ql.internal_exception





# -*- coding: utf-8 -*-
#
# TARGET arch is: []
# WORD_SIZE is: 8
# POINTER_SIZE is: 8
# LONGDOUBLE_SIZE is: 16
#
import ctypes


c_int128 = ctypes.c_ubyte*16
c_uint128 = c_int128
void = None
if ctypes.sizeof(ctypes.c_longdouble) == 16:
    c_long_double_t = ctypes.c_longdouble
else:
    c_long_double_t = ctypes.c_ubyte*16

# if local wordsize is same as target, keep ctypes pointer function.
# required to access _ctypes
import _ctypes
# Emulate a pointer class using the approriate c_int32/c_int64 type
# The new class should have :
# ['__module__', 'from_param', '_type_', '__dict__', '__weakref__', '__doc__']
# but the class should be submitted to a unique instance for each base type
# to that if A == B, POINTER_T(A) == POINTER_T(B)
ctypes._pointer_t_type_cache = {}
def POINTER_T(pointee):
    # a pointer should have the same length as LONG
    fake_ptr_base_type = ctypes.c_uint64 
    # specific case for c_void_p
    if pointee is None: # VOID pointer type. c_void_p.
        pointee = type(None) # ctypes.c_void_p # ctypes.c_ulong
        clsname = 'c_void'
    else:
        clsname = pointee.__name__
    if clsname in ctypes._pointer_t_type_cache:
        return ctypes._pointer_t_type_cache[clsname]
    # make template
    _class = type('LP_%d_%s'%(8, clsname), (fake_ptr_base_type,),{}) 
    ctypes._pointer_t_type_cache[clsname] = _class
    return _class



undefined = ctypes.c_ubyte
ImageBaseOffset32 = ctypes.c_uint32
byte = ctypes.c_ubyte
dword = ctypes.c_uint32
longlong = ctypes.c_int64
qword = ctypes.c_uint64
uchar = ctypes.c_ubyte
uint = ctypes.c_uint32
ulong = ctypes.c_uint64
ulonglong = ctypes.c_uint64
undefined1 = ctypes.c_ubyte
undefined4 = ctypes.c_uint32
undefined8 = ctypes.c_uint64
ushort = ctypes.c_uint16
word = ctypes.c_uint16
class struct_DXE_SERVICES(ctypes.Structure):
    pass

class struct_EFI_TABLE_HEADER(ctypes.Structure):
    _pack_ = True # source:False
    _functions_ = []
    _fields_ = [
    ('Signature', ctypes.c_uint64),
    ('Revision', ctypes.c_uint32),
    ('HeaderSize', ctypes.c_uint32),
    ('CRC32', ctypes.c_uint32),
    ('Reserved', ctypes.c_uint32),
     ]


# values for enumeration 'enum_852'
enum_852__enumvalues = {
    7: 'EfiGcdMemoryTypeMaximum',
    3: 'EfiGcdMemoryTypeMemoryMappedIo',
    6: 'EfiGcdMemoryTypeMoreReliable',
    0: 'EfiGcdMemoryTypeNonExistent',
    4: 'EfiGcdMemoryTypePersistent',
    5: 'EfiGcdMemoryTypePersistentMemory',
    1: 'EfiGcdMemoryTypeReserved',
    2: 'EfiGcdMemoryTypeSystemMemory',
}
EfiGcdMemoryTypeMaximum = 7
EfiGcdMemoryTypeMemoryMappedIo = 3
EfiGcdMemoryTypeMoreReliable = 6
EfiGcdMemoryTypeNonExistent = 0
EfiGcdMemoryTypePersistent = 4
EfiGcdMemoryTypePersistentMemory = 5
EfiGcdMemoryTypeReserved = 1
EfiGcdMemoryTypeSystemMemory = 2
enum_852 = ctypes.c_int # enum
class struct_EFI_GCD_IO_SPACE_DESCRIPTOR(ctypes.Structure):
    pass

class struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR(ctypes.Structure):
    pass

class struct_GUID(ctypes.Structure):
    pass


# values for enumeration 'enum_853'
enum_853__enumvalues = {
    2: 'EfiGcdIoTypeIo',
    3: 'EfiGcdIoTypeMaximum',
    0: 'EfiGcdIoTypeNonExistent',
    1: 'EfiGcdIoTypeReserved',
}
EfiGcdIoTypeIo = 2
EfiGcdIoTypeMaximum = 3
EfiGcdIoTypeNonExistent = 0
EfiGcdIoTypeReserved = 1
enum_853 = ctypes.c_int # enum

# values for enumeration 'enum_854'
enum_854__enumvalues = {
    2: 'EfiGcdAllocateAddress',
    0: 'EfiGcdAllocateAnySearchBottomUp',
    3: 'EfiGcdAllocateAnySearchTopDown',
    1: 'EfiGcdAllocateMaxAddressSearchBottomUp',
    4: 'EfiGcdAllocateMaxAddressSearchTopDown',
    5: 'EfiGcdMaxAllocateType',
}
EfiGcdAllocateAddress = 2
EfiGcdAllocateAnySearchBottomUp = 0
EfiGcdAllocateAnySearchTopDown = 3
EfiGcdAllocateMaxAddressSearchBottomUp = 1
EfiGcdAllocateMaxAddressSearchTopDown = 4
EfiGcdMaxAllocateType = 5
enum_854 = ctypes.c_int # enum
struct_DXE_SERVICES._pack_ = True # source:False
struct_DXE_SERVICES._functions_ = []
struct_DXE_SERVICES._fields_ = [
    ('Hdr', struct_EFI_TABLE_HEADER),
    ('AddMemorySpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_852, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
    ('AllocateMemorySpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_854, enum_852, ctypes.c_uint64, ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(None), POINTER_T(None)))),
    ('FreeMemorySpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
    ('RemoveMemorySpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
    ('GetMemorySpaceDescriptor', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR)))),
    ('SetMemorySpaceAttributes', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
    ('GetMemorySpaceMap', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR))))),
    ('AddIoSpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_853, ctypes.c_uint64, ctypes.c_uint64))),
    ('AllocateIoSpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_854, enum_853, ctypes.c_uint64, ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(None), POINTER_T(None)))),
    ('FreeIoSpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
    ('RemoveIoSpace', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
    ('GetIoSpaceDescriptor', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR)))),
    ('GetIoSpaceMap', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR))))),
    ('Dispatch', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64))),
    ('Schedule', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID)))),
    ('Trust', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID)))),
    ('ProcessFirmwareVolume', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), ctypes.c_uint64, POINTER_T(POINTER_T(None))))),
    ('SetMemorySpaceCapabilities', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
]

struct_DXE_SERVICES._functions_.append(("AddMemorySpace",['ctypes.c_uint64', 'enum_852', 'ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("AllocateMemorySpace",['ctypes.c_uint64', 'enum_854', 'enum_852', 'ctypes.c_uint64', 'ctypes.c_uint64', 'POINTER_T(ctypes.c_uint64)', 'POINTER_T(None)', 'POINTER_T(None)']))
struct_DXE_SERVICES._functions_.append(("FreeMemorySpace",['ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("RemoveMemorySpace",['ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("GetMemorySpaceDescriptor",['ctypes.c_uint64', 'ctypes.c_uint64', 'POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR)']))
struct_DXE_SERVICES._functions_.append(("SetMemorySpaceAttributes",['ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("GetMemorySpaceMap",['ctypes.c_uint64', 'POINTER_T(ctypes.c_uint64)', 'POINTER_T(POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR))']))
struct_DXE_SERVICES._functions_.append(("AddIoSpace",['ctypes.c_uint64', 'enum_853', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("AllocateIoSpace",['ctypes.c_uint64', 'enum_854', 'enum_853', 'ctypes.c_uint64', 'ctypes.c_uint64', 'POINTER_T(ctypes.c_uint64)', 'POINTER_T(None)', 'POINTER_T(None)']))
struct_DXE_SERVICES._functions_.append(("FreeIoSpace",['ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("RemoveIoSpace",['ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("GetIoSpaceDescriptor",['ctypes.c_uint64', 'ctypes.c_uint64', 'POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR)']))
struct_DXE_SERVICES._functions_.append(("GetIoSpaceMap",['ctypes.c_uint64', 'POINTER_T(ctypes.c_uint64)', 'POINTER_T(POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR))']))
struct_DXE_SERVICES._functions_.append(("Dispatch",['ctypes.c_uint64']))
struct_DXE_SERVICES._functions_.append(("Schedule",['ctypes.c_uint64', 'POINTER_T(None)', 'POINTER_T(struct_GUID)']))
struct_DXE_SERVICES._functions_.append(("Trust",['ctypes.c_uint64', 'POINTER_T(None)', 'POINTER_T(struct_GUID)']))
struct_DXE_SERVICES._functions_.append(("ProcessFirmwareVolume",['ctypes.c_uint64', 'POINTER_T(None)', 'ctypes.c_uint64', 'POINTER_T(POINTER_T(None))']))
struct_DXE_SERVICES._functions_.append(("SetMemorySpaceCapabilities",['ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
DXE_SERVICES = struct_DXE_SERVICES
PDXE_SERVICES = POINTER_T(struct_DXE_SERVICES)
EFI_DXE_SERVICES = struct_DXE_SERVICES
EFI_TABLE_HEADER = struct_EFI_TABLE_HEADER
PEFI_TABLE_HEADER = POINTER_T(struct_EFI_TABLE_HEADER)
UINT64 = ctypes.c_uint64
UINTN = ctypes.c_uint64
RETURN_STATUS = ctypes.c_uint64
EFI_STATUS = ctypes.c_uint64
EFI_GCD_MEMORY_TYPE = enum_852
EFI_GCD_MEMORY_TYPE__enumvalues = enum_852__enumvalues
EFI_PHYSICAL_ADDRESS = ctypes.c_uint64
EFI_ADD_MEMORY_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_852, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
EFI_GCD_ALLOCATE_TYPE = enum_854
EFI_GCD_ALLOCATE_TYPE__enumvalues = enum_854__enumvalues
EFI_HANDLE = POINTER_T(None)
EFI_ALLOCATE_MEMORY_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_854, enum_852, ctypes.c_uint64, ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(None), POINTER_T(None)))
EFI_FREE_MEMORY_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
EFI_REMOVE_MEMORY_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR._pack_ = True # source:False
struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR._functions_ = []
struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR._fields_ = [
    ('BaseAddress', ctypes.c_uint64),
    ('Length', ctypes.c_uint64),
    ('Capabilities', ctypes.c_uint64),
    ('Attributes', ctypes.c_uint64),
    ('GcdMemoryType', EFI_GCD_MEMORY_TYPE),
    ('PADDING_0', ctypes.c_ubyte * 4),
    ('ImageHandle', POINTER_T(None)),
    ('DeviceHandle', POINTER_T(None)),
]

EFI_GCD_MEMORY_SPACE_DESCRIPTOR = struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR
PEFI_GCD_MEMORY_SPACE_DESCRIPTOR = POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR)
EFI_GET_MEMORY_SPACE_DESCRIPTOR = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR)))
EFI_SET_MEMORY_SPACE_ATTRIBUTES = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
EFI_GET_MEMORY_SPACE_MAP = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR))))
EFI_GCD_IO_TYPE = enum_853
EFI_GCD_IO_TYPE__enumvalues = enum_853__enumvalues
EFI_ADD_IO_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_853, ctypes.c_uint64, ctypes.c_uint64))
EFI_ALLOCATE_IO_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_854, enum_853, ctypes.c_uint64, ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(None), POINTER_T(None)))
EFI_FREE_IO_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
EFI_REMOVE_IO_SPACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
PEFI_GCD_IO_SPACE_DESCRIPTOR = POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR)
struct_EFI_GCD_IO_SPACE_DESCRIPTOR._pack_ = True # source:False
struct_EFI_GCD_IO_SPACE_DESCRIPTOR._functions_ = []
struct_EFI_GCD_IO_SPACE_DESCRIPTOR._fields_ = [
    ('BaseAddress', ctypes.c_uint64),
    ('Length', ctypes.c_uint64),
    ('GcdIoType', EFI_GCD_IO_TYPE),
    ('PADDING_0', ctypes.c_ubyte * 4),
    ('ImageHandle', POINTER_T(None)),
    ('DeviceHandle', POINTER_T(None)),
]

EFI_GCD_IO_SPACE_DESCRIPTOR = struct_EFI_GCD_IO_SPACE_DESCRIPTOR
EFI_GET_IO_SPACE_DESCRIPTOR = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR)))
EFI_GET_IO_SPACE_MAP = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.c_uint64), POINTER_T(POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR))))
EFI_DISPATCH = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64))
struct_GUID._pack_ = True # source:False
struct_GUID._functions_ = []
struct_GUID._fields_ = [
    ('Data1', ctypes.c_uint32),
    ('Data2', ctypes.c_uint16),
    ('Data3', ctypes.c_uint16),
    ('Data4', ctypes.c_ubyte * 8),
]

GUID = struct_GUID
PGUID = POINTER_T(struct_GUID)
EFI_GUID = struct_GUID
EFI_SCHEDULE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID)))
EFI_TRUST = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID)))
EFI_PROCESS_FIRMWARE_VOLUME = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), ctypes.c_uint64, POINTER_T(POINTER_T(None))))
EFI_SET_MEMORY_SPACE_CAPABILITIES = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
UINT32 = ctypes.c_uint32
UINT16 = ctypes.c_uint16
UINT8 = ctypes.c_ubyte
__all__ = \
    ['DXE_SERVICES', 'EFI_ADD_IO_SPACE', 'EFI_ADD_MEMORY_SPACE',
    'EFI_ALLOCATE_IO_SPACE', 'EFI_ALLOCATE_MEMORY_SPACE',
    'EFI_DISPATCH', 'EFI_DXE_SERVICES', 'EFI_FREE_IO_SPACE',
    'EFI_FREE_MEMORY_SPACE', 'EFI_GCD_ALLOCATE_TYPE',
    'EFI_GCD_ALLOCATE_TYPE__enumvalues',
    'EFI_GCD_IO_SPACE_DESCRIPTOR', 'EFI_GCD_IO_TYPE',
    'EFI_GCD_IO_TYPE__enumvalues', 'EFI_GCD_MEMORY_SPACE_DESCRIPTOR',
    'EFI_GCD_MEMORY_TYPE', 'EFI_GCD_MEMORY_TYPE__enumvalues',
    'EFI_GET_IO_SPACE_DESCRIPTOR', 'EFI_GET_IO_SPACE_MAP',
    'EFI_GET_MEMORY_SPACE_DESCRIPTOR', 'EFI_GET_MEMORY_SPACE_MAP',
    'EFI_GUID', 'EFI_HANDLE', 'EFI_PHYSICAL_ADDRESS',
    'EFI_PROCESS_FIRMWARE_VOLUME', 'EFI_REMOVE_IO_SPACE',
    'EFI_REMOVE_MEMORY_SPACE', 'EFI_SCHEDULE',
    'EFI_SET_MEMORY_SPACE_ATTRIBUTES',
    'EFI_SET_MEMORY_SPACE_CAPABILITIES', 'EFI_STATUS',
    'EFI_TABLE_HEADER', 'EFI_TRUST', 'EfiGcdAllocateAddress',
    'EfiGcdAllocateAnySearchBottomUp',
    'EfiGcdAllocateAnySearchTopDown',
    'EfiGcdAllocateMaxAddressSearchBottomUp',
    'EfiGcdAllocateMaxAddressSearchTopDown', 'EfiGcdIoTypeIo',
    'EfiGcdIoTypeMaximum', 'EfiGcdIoTypeNonExistent',
    'EfiGcdIoTypeReserved', 'EfiGcdMaxAllocateType',
    'EfiGcdMemoryTypeMaximum', 'EfiGcdMemoryTypeMemoryMappedIo',
    'EfiGcdMemoryTypeMoreReliable', 'EfiGcdMemoryTypeNonExistent',
    'EfiGcdMemoryTypePersistent', 'EfiGcdMemoryTypePersistentMemory',
    'EfiGcdMemoryTypeReserved', 'EfiGcdMemoryTypeSystemMemory',
    'GUID', 'ImageBaseOffset32', 'PDXE_SERVICES',
    'PEFI_GCD_IO_SPACE_DESCRIPTOR',
    'PEFI_GCD_MEMORY_SPACE_DESCRIPTOR', 'PEFI_TABLE_HEADER', 'PGUID',
    'RETURN_STATUS', 'UINT16', 'UINT32', 'UINT64', 'UINT8', 'UINTN',
    'byte', 'dword', 'enum_852', 'enum_853', 'enum_854', 'longlong',
    'qword', 'struct_DXE_SERVICES',
    'struct_EFI_GCD_IO_SPACE_DESCRIPTOR',
    'struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR',
    'struct_EFI_TABLE_HEADER', 'struct_GUID', 'uchar', 'uint',
    'ulong', 'ulonglong', 'undefined', 'undefined1', 'undefined4',
    'undefined8', 'ushort', 'word']



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

import os
import uuid
import ntpath
from sys import getsizeof

from qiling.const import *
from qiling.os.const import *

from .registry import RegistryManager
from .clipboard import Clipboard
from .fiber import FiberManager
from .handle import HandleManager, Handle
from .thread import QlWindowsThreadManagement, QlWindowsThread
from .structs import UNICODE_STRING32, UNICODE_STRING64


def ql_x86_windows_hook_mem_error(ql, access, addr, size, value):
    ql.dprint(D_INFO, "[+] ERROR: unmapped memory access at 0x%x" % addr)
    return False


def string_unpack(string):
    return string.decode().split("\x00")[0]


def read_guid(ql, address):
    result = ""
    raw_guid = ql.mem.read(address, 16)
    return uuid.UUID(bytes_le=bytes(raw_guid))


def print_function(ql, passthru, address, function_name, params, ret):
    function_name = function_name.replace('hook_', '')
    if function_name in ("__stdio_common_vfprintf", "__stdio_common_vfwprintf",
                         "printf", "wsprintfW", "sprintf"):
        return
    log = '0x%0.2x: %s(' % (address, function_name)
    for each in params:
        value = params[each]
        if type(value) == str or type(value) == bytearray:
            log += '%s = "%s", ' % (each, value)
        elif type(value) == tuple:
            log += '%s = 0x%x, ' % (each, value[0])
        else:
            log += '%s = 0x%x, ' % (each, value)
    log = log.strip(", ")
    log += ')'
    if ret is not None:
        log += ' = 0x%x' % ret

    if passthru:
        log += ' (PASSTHRU)'

    if ql.output != QL_OUTPUT.DEBUG:
        log = log.partition(" ")[-1]
        ql.nprint(log)
    else:
        ql.dprint(D_INFO, log)


def read_wstring(ql, address):
    result = ""
    char = ql.mem.read(address, 2)
    # decode as utf-16-le only
    while char.decode('utf-16-le', errors="backslashreplace") != "\x00":
        address += 2
        result += char.decode('utf-16-le', errors="backslashreplace")
        char = ql.mem.read(address, 2)

    # We need to remove \x00 inside the string. Compares do not work otherwise
    result = result.replace("\x00", "")
    string_appearance(ql, result)
    return result


def read_wchar(ql, val):
    return val.decode('utf-16-le', errors="backslashreplace")


def read_punicode_string(ql, address):
    if ql.archbit == 64:  # Win64
        ub = ql.mem.read(address, getsizeof(UNICODE_STRING64))
        us = UNICODE_STRING64.from_buffer_copy(ub)
    else:  # Win32
        ub = ql.mem.read(address, getsizeof(UNICODE_STRING32))
        us = UNICODE_STRING32.from_buffer_copy(ub)

    return read_wstring(ql, us.Buffer)


def read_cstring(ql, address):
    result = ""
    char = ql.mem.read(address, 1)
    while char.decode(errors="ignore") != "\x00":
        address += 1
        result += char.decode(errors="ignore")
        char = ql.mem.read(address, 1)
    string_appearance(ql, result)
    return result


def env_dict_to_array(env_dict):
    env_list = []
    for item in env_dict:
        env_list.append(item + "=" + env_dict[item])
    return env_list


def debug_print_stack(ql, num, message=None):
    if message:
        ql.dprint(D_INFO, "========== %s ==========" % message)
        sp = ql.reg.arch_sp
        ql.dprint(D_INFO, hex(sp + ql.pointersize * num) + ": " + hex(ql.stack_read(num * ql.pointersize)))


def is_file_library(string):
    string = string.lower()
    extension = string[-4:]
    return extension in (".dll", ".exe", ".sys", ".drv")


def string_to_hex(string):
    return ":".join("{:02x}".format(ord(c)) for c in string)


def string_appearance(ql, string):
    strings = string.split(" ")
    for string in strings:
        val = ql.os.appeared_strings.get(string, set())
        val.add(ql.os.syscalls_counter)
        ql.os.appeared_strings[string] = val


def printf(ql,
           address,
           fmt,
           params_addr,
           name,
           wstring=False,
           double_pointer=False):
    count = fmt.count("%")
    params = []
    if count > 0:
        for i in range(count):
            # We don't need to mem_read here, otherwise we have a problem with strings, since read_wstring/read_cstring
            #  already take a pointer, and we will have pointer -> pointer -> STRING instead of pointer -> STRING
            params.append(params_addr + i * ql.pointersize, )

        formats = fmt.split("%")[1:]
        index = 0
        for f in formats:
            if f.startswith("s"):
                if wstring:
                    if double_pointer:
                        params[index] = ql.unpack32(
                            ql.mem.read(params[index], ql.pointersize))
                    params[index] = read_wstring(ql, params[index])
                else:
                    params[index] = read_cstring(ql, params[index])
            else:
                # if is not a string, then they are already values!
                pass
            index += 1

        output = '%s(format = %s' % (name, repr(fmt))
        for each in params:
            if type(each) == str:
                output += ', "%s"' % each
            else:
                output += ', 0x%0.2x' % each
        output += ')'
        fmt = fmt.replace("%llx", "%x")
        stdout = fmt % tuple(params)
        output += " = 0x%x" % len(stdout)
    else:
        output = '%s(format = %s) = 0x%x' % (name, repr(fmt), len(fmt))
        stdout = fmt
    ql.nprint(output)
    ql.os.stdout.write(bytes(stdout, 'utf-8'))
    return len(stdout), stdout

def canonical_path(ql, file_path):
    file_path = file_path.replace("C:", ql.rootfs)
    file_path = file_path.replace("\\", os.sep)
    if ql.archbit == 32:
        real_path = os.path.join(ql.rootfs, "Windows", "System32")
        if not os.path.exists(real_path):
            real_path = os.path.join(ql.rootfs, "Windows", "SysWOW64")
    else:
        real_path = os.path.join(ql.rootfs, "Windows", "System32")

    system_path = os.path.join(ql.rootfs, "Windows", "System32")
    if system_path in file_path:
        file_path = file_path.replace(system_path, real_path)
    return file_path

def path_leaf(path):
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)
def find_size_function(ql, func_addr):
    # We have to retrieve the return address position
    code = ql.mem.read(func_addr, 0x100)
    return_procedures = [b"\xc3", b"\xc2", b"\xcb", b"\xca"]
    min_index = min([code.index(return_value) for return_value in return_procedures if return_value in code])
    return min_index



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

"""
This module is intended for general purpose functions that can be used
thoughout the qiling framework
"""
import importlib, logging, os
from .exception import *
from .const import QL_ARCH, QL_ARCH_ALL, QL_OS, QL_OS_ALL, QL_OUTPUT, QL_DEBUGGER, QL_ARCH_32BIT, QL_ARCH_64BIT, QL_ARCH_16BIT
from .const import debugger_map, arch_map, os_map, D_INFO

from unicorn import UcError, UC_ERR_READ_UNMAPPED, UC_ERR_FETCH_UNMAPPED

def catch_KeyboardInterrupt(ql):
    def decorator(func):
        def wrapper(*args, **kw):
            try:
                return func(*args, **kw)
            except BaseException as e:
                from .os.const import THREAD_EVENT_UNEXECPT_EVENT
                ql.os.stop(stop_event=2)
                ql.internal_exception = e
        return wrapper
    return decorator

def ql_get_arch_bits(arch):
    if arch in QL_ARCH_16BIT:
        return 16
    if arch in QL_ARCH_32BIT:
        return 32
    if arch in QL_ARCH_64BIT:
        return 64
    raise QlErrorArch("[!] Invalid Arch")

def ql_is_valid_ostype(ostype):
    if ostype not in QL_OS_ALL:
        return False
    return True

def ql_is_valid_arch(arch):
    if arch not in QL_ARCH_ALL:
        return False
    return True

def loadertype_convert_str(ostype):
    adapter = {
        QL_OS.LINUX: "ELF",
        QL_OS.MACOS: "MACHO",
        QL_OS.FREEBSD: "ELF",
        QL_OS.WINDOWS: "PE",
        QL_OS.UEFI: "PE_UEFI",
        QL_OS.DOS: "DOS",
    }
    return adapter.get(ostype)

def ostype_convert_str(ostype):
    adapter = {}
    adapter.update(os_map)
    adapter = {v: k for k, v in adapter.items()}
    return adapter.get(ostype)

def ostype_convert(ostype):
    # this is for ql.platform
    if ostype == "darwin":
        ostype = "macos"
    adapter = {}
    adapter.update(os_map)
    if ostype in adapter:
        return adapter[ostype]
    # invalid
    return None, None

def arch_convert_str(arch):
    adapter = {}
    adapter.update(arch_map)
    adapter = {v: k for k, v in adapter.items()}
    return adapter.get(arch)

def arch_convert(arch):
    adapter = {}
    adapter.update(arch_map)
    if arch in adapter:
        return adapter[arch]
    # invalid
    return None, None

def output_convert(output):
    adapter = {
        None: QL_OUTPUT.DEFAULT,
        "default": QL_OUTPUT.DEFAULT,
        "disasm": QL_OUTPUT.DISASM,
        "debug": QL_OUTPUT.DEBUG,
        "dump": QL_OUTPUT.DUMP,
    }
    if output in adapter:
        return adapter[output]
    # invalid
    return None, None

def debugger_convert(debugger):
    adapter = {}
    adapter.update(debugger_map)
    if debugger in adapter:
        return adapter[debugger]
    # invalid
    return None, None

def debugger_convert_str(debugger_id):
    adapter = {}
    adapter.update(debugger_map)
    adapter = {v: k for k, v in adapter.items()}
    if debugger_id in adapter:
        return adapter[debugger_id]
    # invalid
    return None, None

def ql_build_module_import_name(module, ostype, arch = None):
    ret_str = "qiling." + module

    ostype_str = ostype
    arch_str = arch

    if type(ostype) is QL_OS:
        ostype_str = ostype_convert_str(ostype)
    
    if ostype_str and "loader" not in ret_str:
        ret_str += "." + ostype_str

    if arch:
        # This is because X86_64 is bundled into X86 in arch
        if module == "arch" and arch == QL_ARCH.X8664:  
            arch_str = "x86"
        elif type(arch) is QL_ARCH:
            arch_str = arch_convert_str(arch)
    else:
        arch_str = ostype_str
        
    ret_str += "." + arch_str
    return ret_str

def ql_get_module_function(module_name, function_name = None):
    try:
        imp_module = importlib.import_module(module_name)
    except:
        raise QlErrorModuleNotFound("[!] Unable to import module %s" % module_name)

    try:
        module_function = getattr(imp_module, function_name)
    except:
        raise QlErrorModuleFunctionNotFound("[!] Unable to import %s from %s" % (function_name, imp_module))

    return module_function

def ql_setup_logger(logger_name=None):
    if logger_name is None: # increasing logger name counter to prevent conflict 
        loggers = logging.root.manager.loggerDict
        _counter = len(loggers)
        logger_name = 'qiling_%s' % _counter

    logger = logging.getLogger(logger_name)
    logger.propagate = False
    logger.setLevel(logging.DEBUG)
    return logger

def ql_setup_logging_env(ql, logger=None):
    if not os.path.exists(ql.log_dir):
        os.makedirs(ql.log_dir, 0o755)

    ql.log_filename = ql.targetname + ql.append          
    ql.log_file = os.path.join(ql.log_dir, ql.log_filename) 

    _logger = ql_setup_logging_stream(ql)

    if ql.log_split == False:
        _logger = ql_setup_logging_file(ql.output, ql.log_file, _logger)

    return _logger


def ql_setup_logging_stream(ql, logger=None):

    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    ch.terminator = ""

    if logger is None:
        logger = ql_setup_logger()

    logger.addHandler(ch)
    return logger

def ql_setup_logging_file(ql_mode, log_file_path, logger=None):

    # setup FileHandler for logging to disk file
    fh = logging.FileHandler('%s.qlog' % log_file_path)
    fh.setLevel(logging.DEBUG)
    
    # use empty character for stirng terminateor by default
    fh.terminator = ""

    if logger is None:
        logger = ql_setup_logger()

    logger.addHandler(fh)
    return logger

def ql_setup_filter(func_names=None):
    class _filter(logging.Filter):
        def __init__(self, func_names):
            super().__init__()
            # accept list or string func_names so you can use it in qltool and programming
            self.filter_list = func_names.strip().split(",") if isinstance(func_names, str) else func_names

        def filter(self, record):
            return any((record.getMessage().startswith(each) for each in self.filter_list))

    class _FalseFilter(logging.Filter):
        def __init__(self):
            super().__init__()
        def filter(self, record):
            return False

    if func_names == False:
        return _FalseFilter()
    else:
        return _filter(func_names)

# verify if emulator returns properly
def verify_ret(ql, err):
    ql.dprint(D_INFO, "Got exception %u: init SP = %x, current SP = %x, PC = %x" %(err.errno, ql.os.init_sp, ql.reg.arch_sp, ql.reg.arch_pc))
    # print("Got exception %u: init SP = %x, current SP = %x, PC = %x" %(err.errno, ql.os.init_sp, self.reg.arch_sp, self.reg.arch_pc))

    ql.os.RUN = False

    # timeout is acceptable in this case
    if err.errno in (UC_ERR_READ_UNMAPPED, UC_ERR_FETCH_UNMAPPED):
        if ql.ostype == QL_OS.MACOS:
            if ql.loader.kext_name:
                # FIXME: Should I push saved RIP before every method callings of IOKit object?
                if ql.os.init_sp == ql.reg.arch_sp - 8:
                    pass
                else:
                    raise
        
        if ql.archtype == QL_ARCH.X8664: # Win64
            if ql.os.init_sp == ql.reg.arch_sp or ql.os.init_sp + 8 == ql.reg.arch_sp or ql.os.init_sp + 0x10 == ql.reg.arch_sp:  # FIXME
                # [+] 0x11626	 c3	  	ret
                # print("OK, stack balanced!")
                pass
            else:
                raise
        else:   # Win32
            if ql.os.init_sp + 12 == ql.reg.arch_sp:   # 12 = 8 + 4
                # [+] 0x114dd	 c2 08 00	  	ret 	8
                pass
            else:
                raise
    else:
        raise        



from qiling.const import *
from qiling.os.const import *
from .const import *
from .utils import *
from .smm_sw_dispatch2_type import *
from .fncc import *

pointer_size = ctypes.sizeof(ctypes.c_void_p)

def free_pointers(ql, address, params):
    ql.os.heap.free(address)
    return EFI_SUCCESS

@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL)
    "DispatchFunction": POINTER, #POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))
    "RegisterContext": POINTER, #POINTER_T(struct_EFI_SMM_SW_REGISTER_CONTEXT)
    "DispatchHandle": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_SMM_SW_DISPATCH2_Register(ql, address, params):
    # Let's save the dispatch params, so they can be triggered if needed. 
    ql.os.smm_dispatch.append(params)
    
@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_SMM_SW_DISPATCH2_PROTOCOL)
    "DispatchHandle": POINTER, #POINTER_T(None)
})
def hook_SMM_SW_DISPATCH2_UnRegister(ql, address, params):
    return EFI_SUCCESS



def install_EFI_SMM_SW_DISPATCH2_PROTOCOL(ql, start_ptr):
    efi_smm_sw_dispatch2_protocol = EFI_SMM_SW_DISPATCH2_PROTOCOL()
    ptr = start_ptr
    pointer_size = 8
    efi_smm_sw_dispatch2_protocol.Register = ptr
    ql.hook_address(hook_SMM_SW_DISPATCH2_Register, ptr)
    ptr += pointer_size
    efi_smm_sw_dispatch2_protocol.UnRegister = ptr
    ql.hook_address(hook_SMM_SW_DISPATCH2_UnRegister, ptr)
    ptr += pointer_size
    return (ptr, efi_smm_sw_dispatch2_protocol)

